# -*- coding: utf-8 -*-
"""
uuid helper
"""
import os
import random
import time
import uuid


def create_uuid():
    """
    Generate random string
    """
    return uuid.uuid4()


def create_short_uuid():
    """
    Generate random integer
    """
    result = str(random.randint(0, 999999999))
    return result


def create_short_timestamp_uuid():
    """
    Generate short uuid plus timestamp
    """
    short_uuid = create_short_uuid()
    return "{}{}".format(short_uuid, time.time())


def report_name(feature, browser_type):
    # name_of_feature.chrome.1495298685509.json
    if '/' in feature:
        feature = feature.split('/')[-1]
    elif os.sep in feature:
        feature = feature.split(os.sep)[-1]
    feature_name = remove_suffix(feature, '.feature')
    millis = int(round(time.time() * 1000))
    return f"{feature_name}.{browser_type}.{millis}.json"


def remove_suffix(s: str, suffix: str) -> str:
    # suffix='' should not call self[:-0].
    if suffix and s.endswith(suffix):
        return s[:-len(suffix)]
    else:
        return s[:]


def remove_prefix(s: str, prefix: str) -> str:
    if s.startswith(prefix):
        return s[len(prefix):]
    else:
        return s[:]



# -*- coding: utf-8 -*-
"""
report generate
"""
import traceback

from flybirds.core.launch_cycle.run_manage import RunManage
from flybirds.report.gen_factory import GenFactory
from flybirds.utils.flybirds_log import logger


class OnGenerate:
    """
    generate event class
    """

    name = "OnGenerate"
    order = 200

    @staticmethod
    def can(context):
        """
        only local need gen report
        """
        run_at = context.get("run_at")
        if run_at is not None and run_at != "local":
            logger.info("not run at local,skip generate html report")
            return False
        else:
            return True
        report_path = context["report_dir_path"]
        gen_type = context["report_format"]
        if report_path is not None and report_path != "" and gen_type is \
                not None and gen_type != "":
            return True
        else:
            logger.info("skip generate html report maybe no report path "
                        "or no report generate type")
            return False

    @staticmethod
    def run(context):
        """
        generate logical
        """
        try:
            logger.info("start generate report")
            report_path = context["report_dir_path"]
            gen_type = context["report_format"]
            platform = context["cur_platform"]
            GenFactory.gen(gen_type, report_path, platform)
        except Exception:
            logger.error(
                f"report task execute error: {traceback.format_exc()}")


# add event to processor
RunManage.join("after_run_processor", OnGenerate, 1)



# -*- coding: utf-8 -*-
"""
Command methods.
"""
import json
import re
import time
import traceback

import flybirds.core.global_resource as gr
import flybirds.utils.flybirds_log as log
from flybirds.core.plugin.plugins.default.screen import BaseScreen
from flybirds.core.driver import ui_driver
from flybirds.core.global_context import GlobalContext


def sleep(context, param):
    time.sleep(float(param))


def screenshot(context):
    step_index = context.cur_step_index - 1
    BaseScreen.screen_link_to_behave(context.scenario, step_index, "screen_")


def ocr(context):
    step_index = context.cur_step_index - 1
    image_path = BaseScreen.screen_link_to_behave(context.scenario, step_index, "screen_", False)
    BaseScreen.image_ocr(image_path)


def change_ocr_lang(context,lang=None):
    """
    change ocr language
    """
    ocr_instance = ui_driver.init_ocr(lang)
    gr.set_value("ocrInstance", ocr_instance)
    context.ocr_instance = ocr_instance
    GlobalContext.ocr_driver_instance = ocr_instance
    log.info("ocr change lang complete with {}".format(lang))


def prev_fail_scenario_relevance(context, param1, param2):
    """
    Related operations for the previous failure scenario
    """
    try:
        print("failed info about param", param2)
        fail_info = gr.get_rerun_info(param2.strip())
        if not (fail_info is None):
            if isinstance(fail_info, str):
                fail_info = json.loads(fail_info)
            if isinstance(fail_info, dict):
                scenario = context.scenario
                step_index = context.cur_step_index - 1

                scenario_uri = "failed function: {}。 senario: {}".format(
                    fail_info["feature_name"], fail_info["scenario_name"]
                )
                scenario_uri = scenario_uri.replace(",", "#")
                data = "embeddingsTags, stepIndex={}, <p>{}</p>".format(
                    step_index, scenario_uri
                )
                scenario.description.append(data)

                if isinstance(fail_info["description"], list):
                    # fail_description = fail_info["description"]
                    for des_item in fail_info["description"]:
                        print("des_item", des_item)
                        if des_item.strip().startswith("embeddingsTags"):
                            if "<image" in des_item and "/screen_" in des_item:
                                continue
                            else:
                                scenario.description.append(
                                    re.sub(
                                        r"stepIndex=\d+",
                                        "stepIndex={}".format(step_index),
                                        des_item,
                                        1,
                                    )
                                )
        else:
            log.warn("not find failed senario info: ", param2)
    except Exception:
        log.warn("rerun failed senario error")
        log.warn(traceback.format_exc())


def img_verify(context, search_image_path):
    """
    verify image exist or not
    """
    step_index = context.cur_step_index - 1
    source_image_path = BaseScreen.screen_link_to_behave(context.scenario, step_index, "screen_", False)
    result = BaseScreen.image_verify(source_image_path, search_image_path)
    return result




# -*- coding: utf-8 -*-
"""
manage user py code
"""
import os
import sys


class ScriptImportManage:
    """
    use user py module code
    """

    def __init__(self):
        sys.path.insert(1, os.path.join(os.getcwd(), "pscript"))
        params_deal_path = os.path.join(
            os.getcwd(), "pscript", "params_deal.py"
        )
        if os.path.exists(params_deal_path):
            params_deal_import_path = "params_deal"
            self.params_deal = __import__(params_deal_import_path)
        custom_operation_path = os.path.join(
            os.getcwd(), "pscript", "custom_handle", "operation.py"
        )

        if os.path.exists(custom_operation_path):
            custom_operation_import_path = "custom_handle.operation"
            operation_module = __import__(custom_operation_import_path)
            self.custom_operation = getattr(operation_module, "operation")
            self.app_operation = getattr(operation_module, "operation")

        if not hasattr(self, "custom_operation") \
                or self.custom_operation is None:
            app_operation_path = os.path.join(
                os.getcwd(), "pscript", "app", "operation.py")
            if os.path.exists(app_operation_path):
                app_operation_import_path = "app.operation"
                app_module = __import__(app_operation_import_path)
                self.app_operation = getattr(app_module, "operation")
                self.custom_operation = getattr(app_module, "operation")

        dsl_hook_path = os.path.join(os.getcwd(), "pscript", "dsl", "hook.py")
        if os.path.exists(dsl_hook_path):
            dsl_hook_import_path = "dsl.hook"
            dsl_module = __import__(dsl_hook_import_path)
            self.dsl_hook = getattr(dsl_module, "hook")
        # tag_provider
        tag_provider_path = os.path.join(
            os.getcwd(), "pscript", "tag_provider.py"
        )
        if os.path.exists(tag_provider_path):
            tag_provider_path_import_path = "tag_provider"
            self.tag_provider = __import__(tag_provider_path_import_path)



#! usr/bin/python
# -*- coding:utf-8 -*-
import math
import cv2


def generate_result(rect, confi):
    """Format the result: Define the image recognition result format."""
    ret = {
        'rect': rect,
        'confidence': confi,
    }
    return ret


def keypoint_distance(kp1, kp2):
    """Find the distance between two keypoints"""
    if isinstance(kp1, cv2.KeyPoint):
        kp1 = kp1.pt
    elif isinstance(kp1, (list, tuple)):
        kp1 = kp1
    else:
        raise ValueError('When kp1 needs keypoint or direct coordinates, kp1={}'.format(kp1))

    if isinstance(kp2, cv2.KeyPoint):
        kp2 = kp2.pt
    elif isinstance(kp2, (list, tuple)):
        kp2 = kp2
    else:
        raise ValueError('When kp2 needs keypoint or direct coordinates, kp1={}'.format(kp2))

    x = kp1[0] - kp2[0]
    y = kp1[1] - kp2[1]
    return math.sqrt((x ** 2) + (y ** 2))


def get_keypoint_from_matches(kp, matches, mode):
    res = []
    if mode == 'query':
        for match in matches:
            res.append(kp[match.queryIdx])
    elif mode == 'train':
        for match in matches:
            res.append(kp[match.trainIdx])

    return res


def _mapping_angle_distance(distance, origin_angle, angle):
    """

    Args:
        distance: distance
        origin_angle: The angle corresponding to the origin
        angle: Rotation angle

    """
    _angle = origin_angle + angle
    _y = distance * math.cos((math.pi * _angle) / 180)
    _x = distance * math.sin((math.pi * _angle) / 180)
    return round(_x, 3), round(_y, 3)


def rectangle_transform(point, size, mapping_point, mapping_size, angle):
    """
    According to the point, find the rectangle vertex coordinates mapped by mapping_point

    Args:
         point: the coordinates of the coordinates in the rectangle
         size: the size of the rectangle (h, w)
         mapping_point: the coordinates of the mapping rectangle
         mapping_size: the size of the mapping rectangle (h, w)
         angle: rotation angle

    Returns:

    """
    h, w = size[0], size[1]
    _h, _w = mapping_size[0], mapping_size[1]

    h_scale = _h / h
    w_scale = _w / w

    tl = keypoint_distance((0, 0), point)  # upper left
    tr = keypoint_distance((w, 0), point)  # top right
    bl = keypoint_distance((0, h), point)  # lower left
    br = keypoint_distance((w, h), point)  # lower right

    A = math.degrees(math.atan2(point[0], point[1]))
    B = math.degrees(math.atan2((w - point[0]), point[1]))
    C = math.degrees(math.atan2(point[0], (h - point[1])))
    D = math.degrees(math.atan2((w - point[0]), (h - point[1])))

    new_tl = _mapping_angle_distance(tl, A, angle=angle)
    new_tl = (-new_tl[0] * w_scale, -new_tl[1] * h_scale)
    new_tl = (mapping_point[0] + new_tl[0], mapping_point[1] + new_tl[1])

    new_tr = _mapping_angle_distance(tr, B, angle=angle)
    new_tr = (new_tr[0] * w_scale, -new_tr[1] * h_scale)
    new_tr = (mapping_point[0] + new_tr[0], mapping_point[1] + new_tr[1])

    new_bl = _mapping_angle_distance(bl, C, angle=angle)
    new_bl = (-new_bl[0] * w_scale, new_bl[1] * h_scale)
    new_bl = (mapping_point[0] + new_bl[0], mapping_point[1] + new_bl[1])

    new_br = _mapping_angle_distance(br, D, angle=angle)
    new_br = (new_br[0] * w_scale, new_br[1] * h_scale)
    new_br = (mapping_point[0] + new_br[0], mapping_point[1] + new_br[1])

    return [new_tl, new_tr, new_bl, new_br]



# -*- coding: utf-8 -*-
"""
This module defines the steps related to the web network request.
"""

from behave import step

from flybirds.core.global_context import GlobalContext as g_Context
from flybirds.utils.dsl_helper import ele_wrap


@step("cache service request [{service}]")
@ele_wrap
def add_request_body(context, service=None):
    """
    cache information of network request(service)

    :param context: step context
    :param service: service request name. (string or None).
    """
    g_Context.step.add_request_body(context, service)


@step("remove service request cache [{service}]")
@ele_wrap
def remove_request_body(context, service=None):
    """
    remove cached information of network request(service)

    :param context: step context
    :param service: service request name. (string or None).
    """
    g_Context.step.remove_request_body(context, service)


@step("remove all service request caches")
def clear_all_request_body(context):
    """
    remove all cached information of network request

    :param context: step context
    """
    g_Context.step.clear_all_request_body(context)


@step("listening service [{service}] bind mockCase[{mock_case_id}]")
@ele_wrap
def add_request_mock(context, service=None, mock_case_id=None):
    """
    Listening to network request(service) and binding mock data to it

    :param context: step context
    :param service: service request name. (string or None).
    :param mock_case_id: unique key for get mock data.
         (String, e.g. a numeric string, or other).
    """
    g_Context.step.add_request_mock(context, service, mock_case_id)


@step("remove service listener [{service}]")
@ele_wrap
def remover_request_mock(context, service=None):
    """
    remove listening and mock data binding to network request(service)

    :param context: step context
    :param service: service request name. (string or None).
    """
    g_Context.step.remover_request_mock(context, service)


@step("remove all service listeners")
def clear_all_request_mock(context):
    """
     remove listening and mock data binding for all services
    """
    g_Context.step.clear_all_request_mock(context)


@step(
    "compare service request [{service}] with json file [{target_data_path}]"
)
@ele_wrap
def request_compare(context, service=None, target_data_path=None):
    """
    compare and verify the request's post body with the data of the target path

    :param context: step context
    :param service: service request name. (string or None).
    :param target_data_path: path of target data to be compared
    """
    g_Context.step.request_compare_from_path(context, service,
                                             target_data_path)


@step(
    "compare service non-json request [{service}] with non-json "
    "file [{target_data_path}]")
@ele_wrap
def request_query_str_compare(context, service=None, target_data_path=None):
    """
    compare and verify the post body of a non-json type request with the
    data in the target path

    :param context: step context
    :param service: service request name. (string or None).
    :param target_data_path: path of target data to be compared
    """
    g_Context.step.request_query_str_compare_from_path(context, service,
                                                       target_data_path)


@step(
    "service request [{service}] request parameter [{target_json_path}] "
    "is [{expect_value}]")
@ele_wrap
def request_compare_value(context, service=None, target_json_path=None,
                          expect_value=None):
    """
    compare and verify the request parameter of the service request with
     the expected value

    :param context: step context
    :param service: service request name. (string or None).
    :param target_json_path: json path of request parameter
    :param expect_value: expected value
    """
    g_Context.step.request_compare_value(context, service, target_json_path,
                                         expect_value)



# -*- coding: utf-8 -*-
"""
Jump to page
"""
import flybirds.core.global_resource as gr
import flybirds.utils.dsl_helper as dsl_helper
import flybirds.utils.snippet as snippet


def jump_to_page(context, param):
    param_dict = dsl_helper.params_to_dic(param, "pageName")

    page_name = param_dict["pageName"]
    schema_url = gr.get_page_schema_url(page_name)

    schema_goto_module = gr.get_value("projectScript").custom_operation
    deal_method = None
    params_deal_module = None
    if "dealMethod" in param_dict.keys():
        deal_method = param_dict["dealMethod"]
        params_deal_module = gr.get_value("projectScript").params_deal

    snippet.schema_goto(
        page_name,
        schema_url,
        schema_goto_module,
        deal_method,
        params_deal_module,
    )



# -*- coding: utf-8 -*-
"""
Element Snapshots API
"""
import time

import flybirds.core.plugin.plugins.default.ui_driver.poco.findsnap as findsnap
import flybirds.utils.verify_helper as verify_helper
from flybirds.core.exceptions import FlybirdVerifyException


def find_ele_by_snap(poco, config, optional):
    """
    determine whether the element exists within the specified time
    """
    timeout = optional["timeout"]
    current_wait_second = 1
    find_success = False

    while timeout > 0:
        create_success = False
        try:
            is_need_refresh = findsnap.get_refresh_status()
            if is_need_refresh:
                findsnap.refresh_snap()
            poco_resoure = findsnap.get_snap()
            if not (poco_resoure is None):
                create_success = True
                search_time = current_wait_second
                if search_time > 3:
                    search_time = 3
                find_result = findsnap.snap_find(poco_resoure, config)
                find_success = not (find_result is None)
                if find_success:
                    break
            else:
                time.sleep(current_wait_second)
        except Exception:
            if not create_success:
                time.sleep(current_wait_second)
        if current_wait_second > 3:
            time.sleep(current_wait_second - 3)

        timeout -= current_wait_second
        current_wait_second += 2
        findsnap.refresh_snap()
    if not find_success:
        message = "during {}s time, not find {} in page".format(
            optional["timeout"], config
        )
        raise FlybirdVerifyException(message)


def verify_ele_by_snap(poco, config, optional):
    """
    use snapshots to judge element text
    """
    timeout = optional["timeout"]
    current_wait_second = 1
    find_success = False
    expect_text = config.get("expect_text")
    expect_str = expect_text.strip().replace(u"\u200b", "")
    target_str = None
    while timeout > 0:
        create_success = False
        try:
            is_need_refresh = findsnap.get_refresh_status()
            if is_need_refresh:
                findsnap.refresh_snap()
            poco_resoure = findsnap.get_snap()
            if not (poco_resoure is None):
                create_success = True
                search_time = current_wait_second
                if search_time > 3:
                    search_time = 3
                find_result = findsnap.snap_find(poco_resoure, config)
                print("find_result", find_result)
                find_success = not (find_result is None)
                if not (find_result is None):
                    target_str = (
                        (find_result[0].get("payload").get("text") or "")
                        .strip()
                        .replace(u"\u200b", "")
                    )
                    find_success = target_str == expect_str
                    if find_success:
                        break
            else:
                print("No snapshot information is obtained")
                time.sleep(current_wait_second)
        except Exception:
            if not create_success:
                time.sleep(current_wait_second)
        if current_wait_second > 3:
            time.sleep(current_wait_second - 3)

        timeout -= current_wait_second
        current_wait_second += 1
        findsnap.refresh_snap()
    if not find_success:
        verify_helper.text_equal(expect_str, target_str)



#! usr/bin/python
# -*- coding:utf-8 -*-
""" opencv matchTemplate"""
import warnings
import cv2
import numpy as np
from baseImage import Image, Rect
from baseImage.constant import Place

from .exceptions import MatchResultError, InputImageError
from .utils import generate_result


class MatchTemplate(object):
    METHOD_NAME = 'tpl'
    Dtype = np.uint8
    Place = (Place.Ndarray, )

    def __init__(self, threshold=0.8, rgb=True):
        """
        init

        Args:
             threshold: recognition threshold (0~1)
             rgb: whether to use the rgb channel for verification
        """
        assert 0 <= threshold <= 1, 'threshold value between 0 and 1'

        self.threshold = threshold
        self.rgb = rgb
        self.matcher = cv2.matchTemplate

    def find_best_result(self, im_source, im_search, threshold=None, rgb=None):
        """
        Template matching, return the range with the highest
        matching degree and greater than the threshold

        Args:
             im_source: the image to be matched
             im_search: image template
             threshold: recognition threshold (0~1)
             rgb: whether to use the rgb channel for verification

        Returns:
            generate_result
        """
        threshold = threshold or self.threshold
        rgb = rgb or self.rgb

        im_source, im_search = self.input_image_check(im_source, im_search)
        if im_source.channels == 1:
            rgb = False

        result = self._get_template_result_matrix(im_source=im_source, im_search=im_search)
        # Find the best match

        min_val, max_val, min_loc, max_loc = self.minMaxLoc(result.data)

        h, w = im_search.size
        # Seek credibility
        crop_rect = Rect(max_loc[0], max_loc[1], w, h)

        confidence = self.cal_confidence(im_source, im_search, crop_rect, max_val, rgb)
        # Returns None if the confidence is less than the threshold
        if confidence < (threshold or self.threshold):
            return None
        x, y = max_loc
        rect = Rect(x=x, y=y, width=w, height=h)
        return generate_result(rect, confidence)

    def find_all_results(self, im_source, im_search, threshold=None, rgb=None, max_count=10):
        """
        Template matching, return the range with matching degree greater
        than the threshold, and the maximum number does not exceed max_count

        Args:
             im_source: the image to be matched
             im_search: image template
             threshold:: recognition threshold (0~1)
             rgb: whether to use the rgb channel for verification
             max_count: maximum number of matches

        Returns:

        """
        threshold = threshold or self.threshold
        rgb = rgb or self.rgb

        im_source, im_search = self.input_image_check(im_source, im_search)
        if im_source.channels == 1:
            rgb = False

        result = self._get_template_result_matrix(im_source=im_source, im_search=im_search)
        results = []
        # Find the best match
        h, w = im_search.size
        while True:
            min_val, max_val, min_loc, max_loc = self.minMaxLoc(result.data)
            img_crop = im_source.crop(Rect(max_loc[0], max_loc[1], w, h))
            confidence = self._get_confidence_from_matrix(img_crop, im_search, max_val=max_val, rgb=rgb)
            x, y = max_loc
            rect = Rect(x, y, w, h)

            if (confidence < (threshold or self.threshold)) or len(results) >= max_count:
                break
            results.append(generate_result(rect, confidence))
            result.rectangle(rect=Rect(int(max_loc[0] - w / 2), int(max_loc[1] - h / 2), w, h), color=(0, 0, 0), thickness=-1)

        return results if results else None

    def _get_template_result_matrix(self, im_source, im_search):
        """Get the result matrix of template matching."""
        if im_source.channels == 3:
            i_gray = im_source.cvtColor(cv2.COLOR_BGR2GRAY).data
            s_gray = im_search.cvtColor(cv2.COLOR_BGR2GRAY).data
        else:
            i_gray = im_source.data
            s_gray = im_search.data

        result = self.match(i_gray, s_gray)
        result = Image(data=result, dtype=np.float32, clone=False, place=im_source.place)
        return result

    def input_image_check(self, im_source, im_search):
        im_source = self._image_check(im_source)
        im_search = self._image_check(im_search)

        if im_source.place != im_search.place:
            raise InputImageError('image type must be same, source={}, search={}'.format(im_source.place, im_search.place))
        elif im_source.dtype != im_search.dtype:
            raise InputImageError('image data type must be same, source={}, search={}'.format(im_source.dtype, im_search.dtype))
        elif im_source.channels != im_search.channels:
            raise InputImageError('image channel must be same, source={}, search={}'.format(im_source.channels, im_search.channels))

        if im_source.place == Place.UMat:
            warnings.warn('Umat has error,will clone new image with np.ndarray '
                          '(https://github.com/opencv/opencv/issues/21788)')
            im_source = Image(im_source, place=Place.Ndarray, dtype=im_source.dtype)
            im_search = Image(im_search, place=Place.Ndarray, dtype=im_search.dtype)

        return im_source, im_search

    def _image_check(self, data):
        if not isinstance(data, Image):
            data = Image(data, dtype=self.Dtype)

        if data.place not in self.Place:
            raise TypeError('Image type must be(Place.UMat, Place.Ndarray)')
        return data

    @staticmethod
    def minMaxLoc(result):
        return cv2.minMaxLoc(result)

    def match(self, img1, img2):
        return self.matcher(img1, img2, cv2.TM_CCOEFF_NORMED)

    def cal_confidence(self, im_source, im_search, crop_rect, max_val, rgb):
        """
        Scale the screenshot and the recognition result
        to the same size, and calculate the reliability

        Args:
             im_source: the image to be matched
             im_search: image template
             crop_rect: The area that needs to be intercepted in im_source
             max_val: the maximum value obtained by matchTemplate
             rgb: whether to use the rgb channel for verification

        Returns:
            float: credibility(0~1)
        """
        try:
            target_img = im_source.crop(crop_rect)
        except OverflowError:
            raise MatchResultError(f"Target area({crop_rect}) out of screen{im_source.size}")
        confidence = self._get_confidence_from_matrix(target_img, im_search, max_val, rgb)
        return confidence

    def cal_rgb_confidence(self, im_source, im_search):
        """
        Calculate the confidence of two picture rgb three-channel

        Args:
             im_source: the image to be matched
             im_search: image template

        Returns:
            float: minimum confidence
        """
        # im_search = im_search.copyMakeBorder(10, 10, 10, 10, cv2.BORDER_REPLICATE)
        #
        # img_src_hsv = im_source.cvtColor(cv2.COLOR_BGR2HSV)
        # img_sch_hsv = im_search.cvtColor(cv2.COLOR_BGR2HSV)

        src_split = im_source.split()
        sch_split = im_search.split()

        # Calculate the confidence of the BGR three channels and store it in bgr_confidence:
        bgr_confidence = [0, 0, 0]
        for i in range(3):
            res_temp = self.match(sch_split[i], src_split[i])
            min_val, max_val, min_loc, max_loc = self.minMaxLoc(res_temp)
            bgr_confidence[i] = max_val

        return min(bgr_confidence)

    def cal_ccoeff_confidence(self, im_source, im_search):
        if im_source.channels == 3:
            img_src_gray = im_source.cvtColor(cv2.COLOR_BGR2GRAY).data
        else:
            img_src_gray = im_source.data

        if im_search.channels == 3:
            img_sch_gray = im_search.cvtColor(cv2.COLOR_BGR2GRAY).data
        else:
            img_sch_gray = im_search.data

        res_temp = self.match(img_sch_gray, img_src_gray)
        min_val, max_val, min_loc, max_loc = self.minMaxLoc(res_temp)
        return max_val

    def _get_confidence_from_matrix(self, img_crop, im_search, max_val, rgb):
        """Find confidence from the result matrix."""
        # Seek credibility:
        if rgb:
            # If there is color verification, perform
            # BGR three-channel verification on the target area:
            confidence = self.cal_rgb_confidence(img_crop, im_search)
        else:
            confidence = max_val
        return confidence





# -*- coding: utf-8 -*-
# @Time : 2022/3/7 19:18
# @Author : hyx
# @File : page.py
# @desc : web page implement
import json
import time
from urllib.parse import urlparse

import flybirds.core.global_resource as gr
from flybirds.core.global_context import GlobalContext
import flybirds.utils.flybirds_log as log
import flybirds.utils.verify_helper as verify_helper
from flybirds.core.plugin.plugins.default.web.interception import \
    get_case_response_body
from flybirds.utils import dsl_helper
from flybirds.utils.dsl_helper import is_number

__open__ = ["Page"]


class Page:
    """Web Page Class"""

    name = "web_page"
    instantiation_timing = "plugin"

    def __init__(self):
        page, context = self.init_page()
        self.page = page
        self.context = context

    @staticmethod
    def init_page():
        context = gr.get_value("browser_context")
        if context is None or gr.get_web_info_value("browserExit") is None \
                or gr.get_web_info_value("browserExit") is True:
            context = Page.new_browser_context()

        page = context.new_page()
        request_interception = gr.get_web_info_value("request_interception",
                                                     True)
        if request_interception:
            page.route("**/*", handle_route)
            # request listening events
            page.on("request", handle_request)
        page.on("console", handle_page_error)

        ele_wait_time = gr.get_frame_config_value("wait_ele_timeout", 30)
        page_render_timeout = gr.get_frame_config_value("page_render_timeout",
                                                        30)
        page.set_default_timeout(float(ele_wait_time) * 1000)
        page.set_default_navigation_timeout(float(page_render_timeout) * 1000)
        return page, context

    @staticmethod
    def new_browser_context():
        browser = gr.get_value('browser')

        operation_module = gr.get_value("projectScript").custom_operation
        context = None
        if operation_module is not None and \
                hasattr(operation_module, "create_browser_context"):
            create_browser_context = getattr(operation_module,
                                             "create_browser_context")
            if create_browser_context is not None:
                context = create_browser_context(browser)
                if context is not None:
                    log.info(
                        '[new_browser_context] successfully get BrowserContext '
                        'from custom operation')
                    return context

        optional_config = Page.get_web_option_config()
        if optional_config is not None:
            context = browser.new_context(**optional_config,
                                          record_video_dir="videos",
                                          ignore_https_errors=True)
        else:
            context = browser.new_context(record_video_dir="videos",
                                          ignore_https_errors=True)

        # add user custom cookies into browser context
        user_cookie = GlobalContext.get_global_cache("cookies")
        if user_cookie is not None:
            context.clear_cookies()
            context.add_cookies(cookies=user_cookie)
            log.info(f"this is user cookies: {context.cookies()}")
        else:
            log.info(f"user cookies is None")

        return context

    @staticmethod
    def get_web_option_config():
        emulated_device = None
        user_agent = None
        viewport = None
        device_scale_factor = None
        locale = None
        timezone = None
        permissions = None
        geolocation = None
        has_touch = None
        default_browser_type = None
        gl_dict = {}
        if gr.get_web_info_value("emulated_device") is not None:
            playwright = gr.get_value("playwright")
            emulated_device = playwright.devices[
                gr.get_web_info_value("emulated_device")]
        if gr.get_web_info_value("user_agent") is not None:
            user_agent = gr.get_web_info_value("user_agent")
        if gr.get_web_info_value("locale") is not None:
            locale = gr.get_web_info_value("locale")
        if gr.get_web_info_value("timezone") is not None:
            timezone = gr.get_web_info_value("timezone")
        if gr.get_web_info_value("permissions") is not None:
            permissions = gr.get_web_info_value("permissions")
        if gr.get_web_info_value("geolocation") is not None:
            geolocation = gr.get_web_info_value("geolocation")
        if gr.get_web_info_value("width") is not None and gr.get_web_info_value(
                "height"):
            viewport = {'width': gr.get_web_info_value("width"),
                        'height': gr.get_web_info_value(
                            "height")}
        if gr.get_web_info_value("device_scale_factor") is not None:
            device_scale_factor = gr.get_web_info_value("device_scale_factor")

        if gr.get_web_info_value("has_touch") is not None:
            has_touch = gr.get_web_info_value("has_touch")

        if gr.get_web_info_value("default_browser_type") is not None:
            default_browser_type = gr.get_web_info_value("default_browser_type")

        if user_agent is not None:
            gl_dict["user_agent"] = user_agent
        if viewport is not None:
            gl_dict["viewport"] = viewport
        if locale is not None:
            gl_dict["locale"] = locale
        if timezone is not None:
            gl_dict["timezone_id"] = timezone
        if geolocation is not None:
            gl_dict["geolocation"] = geolocation
            if permissions is None:
                permissions = ["geolocation"]
            else:
                if permissions.index("geolocation") < 0:
                    permissions.append("geolocation")
        if permissions is not None:
            gl_dict["permissions"] = permissions
        if device_scale_factor is not None:
            gl_dict["device_scale_factor"] = device_scale_factor
        if has_touch is not None:
            gl_dict["hasTouch"] = has_touch
        if default_browser_type is not None:
            gl_dict["default_browser_type"] = default_browser_type

        if emulated_device is not None:
            gl_dict.update(emulated_device)

        if gl_dict is not None and len(gl_dict) > 0:
            return gl_dict
        else:
            return None

    def navigate(self, context, param):
        operation_module = gr.get_value("projectScript").custom_operation
        page_url = None
        if hasattr(operation_module, "get_page_url"):
            get_page_url = getattr(operation_module,
                                   "get_page_url")
            page_url = get_page_url(param)

        if page_url is not None:
            log.info('[get_page_url] successfully get page_url_value '
                     'from custom operation')
            self.page.goto(page_url)
            return

        param_dict = dsl_helper.params_to_dic(param, "urlKey")
        url_key = param_dict["urlKey"]
        schema_url_value = gr.get_page_schema_url(url_key)
        self.page.goto(schema_url_value)

    def return_pre_page(self, context):
        self.page.go_back()

    def sleep(self, context, param):
        if is_number(param):
            self.page.wait_for_timeout(float(param) * 1000)
        else:
            log.warn("default wait for timeout!")
            self.page.wait_for_timeout(3 * 1000)

    def cur_page_equal(self, context, param):
        cur_url = self.page.url.split('?')[0]
        if param.startswith(("http", "https")):
            target_url = param.split('?')[0]
        else:
            schema_url = gr.get_page_schema_url(param)
            target_url = schema_url
        verify_helper.text_equal(target_url, cur_url)


def handle_page_error(msg):
    if hasattr(msg, "type") and msg.type is not None:
        need_log = False
        if msg.type.lower() == "warn":
            need_log = True
        if msg.type.lower() == "error":
            need_log = True
        if need_log:
            if hasattr(msg, "text"):
                log.info(
                    f"=====================page console==================:\n {msg.text}")


def handle_request(request):
    # interception request handle
    parsed_uri = urlparse(request.url)
    operation = parsed_uri.path.split('/')[-1]
    if operation is not None:
        interception_request = gr.get_value('interceptionRequest')
        request_body = interception_request.get(operation)

        if request_body is not None:
            log.info(
                f'[handle_request] start cache service：{operation}')
            current_request_info = {'postData': request.post_data,
                                    'url': request.url,
                                    'updateTimeStamp': int(
                                        round(time.time() * 1000))}
            interception_request[operation] = current_request_info
            gr.set_value("interceptionRequest", interception_request)


def handle_route(route):
    abort_domain_list = gr.get_web_info_value("abort_domain_list", [])
    parsed_uri = urlparse(route.request.url)
    domain = parsed_uri.hostname
    if abort_domain_list and domain in abort_domain_list:
        route.abort()
        return

    resource_type = route.request.resource_type
    if resource_type != 'fetch' and resource_type != 'xhr':
        route.continue_()
        return

    # mock response data
    operation = parsed_uri.path.split('/')[-1]
    mock_case_id = None
    if operation is not None:
        interception_values = gr.get_value('interceptionValues')
        mock_case_id = interception_values.get(operation)
    if mock_case_id:
        mock_body = get_case_response_body(mock_case_id)
        if mock_body:
            if not isinstance(mock_body, str):
                mock_body = json.dumps(mock_body)
            route.fulfill(status=200,
                          content_type="application/json;charset=utf-8",
                          body=mock_body)
    else:
        route.continue_()



# -*- coding: utf-8 -*-
"""
Poco element click
"""
import flybirds.core.global_resource as gr
import flybirds.core.plugin.plugins.default.ui_driver.poco.findsnap as findsnap
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_ele as poco_ele
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_manage as pm
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_position as pi
from flybirds.core.plugin.plugins.default.ui_driver.poco import poco_text
from flybirds.core.exceptions import PositionNotChangeException


def air_bdd_click(
        poco,
        select_dsl_str,
        optional,
        verify_dsl_str=None,
        verify_optional=None,
        verify_action=None,
):
    """
    click on the element,
    optional parameters to determine whether the clicked page is rendered
    :param poco:
    :param select_dsl_str:
    :param optional: Optional parameters
    :param verify_dsl_str:
    :param verify_optional:
    :param verify_action:
    :return:
    """

    poco_ele.wait_exists(poco, select_dsl_str, optional)
    poco_object = pm.create_poco_object_by_dsl(
        poco, select_dsl_str, optional
    )

    o_position = None
    o_text = None
    if (not (verify_action is None)) and verify_action == "position":
        verify_poco_object = pm.create_poco_object_by_dsl(
            poco, verify_dsl_str, verify_optional
        )
        o_position = verify_poco_object.get_position()
    elif (not (verify_action is None)) and verify_action == "text":
        verify_poco_object = pm.create_poco_object_by_dsl(
            poco, verify_dsl_str, verify_optional
        )
        o_text = verify_poco_object.get_text()

    poco_object.click()
    if gr.get_frame_config_value("use_snap", False):
        # findsnap.refresh_snap()
        findsnap.fix_refresh_status(True)
    if not (verify_dsl_str is None):
        verify_click_end(
            poco,
            verify_dsl_str,
            verify_optional,
            verify_action,
            o_position,
            o_text,
        )


def verify_click_end(poco, verify_dsl_str, verify_optional, verify_action,
                     o_position, o_text
                     ):
    """
    determine whether the rendering of the click effect is completed according
    to the movement of the element,the disappearance of the element or the text
    :param poco:
    :param verify_dsl_str:
    :param verify_optional:
    :param o_position:
    :param verify_action:
    :return:
    """
    if verify_action == "position":
        pos_change = pi.position_change(
            poco, verify_dsl_str, verify_optional, o_position
        )
        if not pos_change:
            raise PositionNotChangeException(
                "during time={} the position of selector={}"
                " not changed".format(
                    verify_optional["timeout"], verify_dsl_str
                )
            )
    elif verify_action == "text":
        pos_change = poco_text.text_change(
            poco, verify_dsl_str, verify_optional, o_text
        )
        if not pos_change:
            raise PositionNotChangeException(
                "during time={} the text of selector={} not changed".format(
                    verify_optional["timeout"], verify_dsl_str
                )
            )
    elif verify_action == "appear":
        poco_ele.wait_exists(poco, verify_dsl_str, verify_optional)
    elif verify_action == "disappear":
        poco_ele.wait_disappear(poco, verify_dsl_str, verify_optional)



#! usr/bin/python
# -*- coding:utf-8 -*-
import cv2
import numpy as np
from baseImage.constant import Place

from .base import BaseKeypoint


class SIFT(BaseKeypoint):
    FLANN_INDEX_KDTREE = 0
    METHOD_NAME = 'SIFT'
    Dtype = np.uint8
    Place = (Place.UMat, Place.Ndarray)

    def create_matcher(self, **kwargs) -> cv2.DescriptorMatcher:
        """
        Create a feature point matcher

        Returns:
            cv2.FlannBasedMatcher
        """
        index_params = {'algorithm': self.FLANN_INDEX_KDTREE, 'tree': 5}
        # Specifies the number of recursive traversals.
        # The higher the value, the more accurate the result, but the more time it takes
        search_params = {'checks': 50}
        matcher = cv2.FlannBasedMatcher(index_params, search_params)
        return matcher

    def create_detector(self, **kwargs) -> cv2.SIFT:
        nfeatures = kwargs.get('nfeatures', 0)
        nOctaveLayers = kwargs.get('nOctaveLayers', 3)
        contrastThreshold = kwargs.get('contrastThreshold', 0.04)
        edgeThreshold = kwargs.get('edgeThreshold', 10)
        sigma = kwargs.get('sigma', 1.6)

        detector = cv2.SIFT_create(nfeatures=nfeatures, nOctaveLayers=nOctaveLayers, contrastThreshold=contrastThreshold,
                                   edgeThreshold=edgeThreshold, sigma=sigma)
        return detector



# -*- coding: utf-8 -*-
"""
Convert poco object
"""
import flybirds.core.plugin.plugins.default.ui_driver.poco.parse_selector \
    as msd
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_selector as pc
import flybirds.utils.snippet as snippet
from flybirds.core.global_context import GlobalContext as g_Context
from flybirds.utils import language_helper as lan


def create_path_poco(poco, path, first_selector="name"):
    """
    convert the path of the ui element described in natural language into
    a poco object.
    """
    multi_list = path.split("->")
    poco_object = None

    # get current language
    language = g_Context.get_current_language()

    for index in range(len(multi_list)):
        cur_selector = multi_list[index].strip()
        if index == 0:
            poco_object = pc.create_poco_object(
                poco,
                msd.create_multi_selector(
                    cur_selector, first_selector
                ),
            )
        else:
            if cur_selector == lan.parse_glb_str("parent", language):
                poco_object = pc.create_parent(poco_object)
            else:
                target_index = 1
                select_dic = {}
                hierarchy_tag = cur_selector
                if cur_selector.startswith(
                        lan.parse_glb_str("rank", language)):
                    target_index = snippet.chose_first_number(cur_selector)
                else:
                    cur_index_selectors = cur_selector.split(" ", 1)
                    target_index = snippet.chose_first_number(
                        cur_index_selectors[0].strip()
                    )
                    select_dic = msd.create_multi_selector(
                        cur_index_selectors[1].strip()
                    )
                    hierarchy_tag = cur_index_selectors[0].strip()
                if lan.parse_glb_str("children", language) in hierarchy_tag:
                    if target_index == 1:
                        poco_object = pc.create_first_child(
                            poco_object, select_dic
                        )
                    else:
                        poco_object = pc.select_child(
                            poco_object, target_index, select_dic
                        )
                if lan.parse_glb_str("sibling", language) in hierarchy_tag:
                    if target_index == 1:
                        poco_object = pc.create_first_sibling(
                            poco_object, select_dic
                        )
                    else:
                        poco_object = pc.select_sibling(
                            poco_object, target_index, select_dic
                        )
                if lan.parse_glb_str("offsprings", language) in hierarchy_tag:
                    if target_index == 1:
                        poco_object = pc.create_first_offspring(
                            poco_object, select_dic
                        )
                    else:
                        poco_object = pc.select_offspring(
                            poco_object, target_index, select_dic
                        )
    return poco_object



# -*- coding: utf-8 -*-
"""
web UI driver core api implement.
"""

from playwright.sync_api import sync_playwright

import flybirds.core.global_resource as gr
import flybirds.utils.flybirds_log as log

__open__ = ["UIDriver"]


class UIDriver:
    """web UI driver Class"""

    name = "web_ui_driver"

    @staticmethod
    def init_driver():
        try:
            play_wright = sync_playwright().start()
            gr.set_value("playwright", play_wright)
            browser_val = gr.get_value("cur_browser")
            browser_type = getattr(play_wright, browser_val)
            headless = gr.get_web_info_value("headless", True)
            browser = browser_type.launch(headless=headless)
            log.info(f"Init browser success! browser_type:[{browser_type}]")
            gr.set_value("browser", browser)

            return play_wright, browser
        except Exception as e:
            log.error('Init browser has error! Error msg is:', str(e))

    @staticmethod
    def close_driver():
        play_wright = gr.get_value('playwright')
        browser = gr.get_value('browser')
        page_obj = gr.get_value('plugin_page')

        if page_obj and hasattr(page_obj, 'context'):
            page_obj.context.close()
        if browser:
            browser.close()
        if play_wright:
            play_wright.stop()



# -*- coding: utf-8 -*-
"""
This module is used to create cli project.
"""
import os
import platform
import shutil

import typer

import flybirds.template as template
import flybirds.utils.flybirds_log as log
from flybirds.utils.file_helper import get_files_from_dir, \
    get_paths_from_dir, \
    replace_file_content, update, update_json_data
from flybirds.utils.pkg_helper import find_package_base_path


def create_demo():
    """
    Create project cli demo
    """
    typer.secho(
        "Welcome to flybirds cli. Please enter any information to continue.",
        fg=typer.colors.MAGENTA,
    )
    user_dict = {
        'device_id': "127.0.0.1:8200",
        'package_name': "ctrip.android.view",
        'web_driver_agent': "com.fd.test.WebDriverAgentLib.xctrunner",
        'headless': True,
        'browser_type': ['chromium']
    }
    project_name = typer.prompt("Please input your project name>>")
    user_dict['project_name'] = project_name
    platform_start = "Please input your test platform? "
    platform_ending = typer.style("(Android/IOS/Web)", fg=typer.colors.CYAN,
                                  bold=True)
    p_message = platform_start + platform_ending
    test_platform = typer.prompt(p_message)
    if test_platform is None or test_platform.strip().lower() not in [
        'android', 'ios', 'web']:
        test_platform = 'android'
    test_platform = test_platform.strip().lower()
    user_dict['test_platform'] = test_platform

    if test_platform in ['android', 'ios']:
        if test_platform == 'ios':
            user_dict['package_name'] = "com.ctrip.inner.wireless"
            is_bundle = typer.confirm(
                "Do you want to configure your webDriverAgent now?"
                "(this step can be skipped)")
            if is_bundle:
                web_driver_agent = typer.prompt(
                    "Please input your Bundle ID of"
                    " webDriverAgent?")
                user_dict['web_driver_agent'] = web_driver_agent
            else:
                typer.secho(
                    "You can configure your  Bundle ID of webDriverAgent later"
                    " in the project's"
                    " flybirds_config.json file.", fg=typer.colors.YELLOW)
        connect_device = typer.confirm(
            "Do you want to configure your deviceId now?"
            "(this step can be skipped)")
        if connect_device:
            device_id = typer.prompt("Please input your deviceId?")
            user_dict['device_id'] = device_id
        else:
            typer.secho(
                "You can configure your deviceId later in the project's"
                " flybirds_config.json file.", fg=typer.colors.YELLOW)

        if_package = typer.confirm(
            "Do you want to configure your packageName now?"
            "(this step can be skipped)")
        if if_package:
            package_name = typer.prompt(
                "Please input your packageName?(You can use"
                " the ADB command to view your package name"
                ", such as: adb shell pm list packages |"
                " findstr 'trip')"
            )
            user_dict['package_name'] = package_name
        else:
            typer.secho(
                "You can configure your packageName later in the project's"
                " flybirds_config.json file.", fg=typer.colors.YELLOW)
    if test_platform == 'web':
        message_start = "Please enter the number represented by the " \
                        "browserType you want to test? Multiple browsers " \
                        "are separated by commas(,)."
        ending = typer.style("(1:chromium  2:firefox  3:webkit)",
                             fg=typer.colors.CYAN, bold=True)
        message = message_start + ending
        out_index = typer.prompt(message)
        index_arr = out_index.strip().split(',')
        browser_dict = {
            '1': "chromium",
            '2': "firefox",
            '3': "webkit"
        }
        browser_types = []
        [browser_types.append(browser_dict.get(i)) for i in index_arr if
         i in browser_dict.keys()]
        # add default value
        if len(browser_types) < 1:
            browser_types.append('chromium')
        user_dict['browser_type'] = browser_types
        headless = typer.confirm(
            "Do you want to launch browser in headless mode?")
        user_dict['headless'] = headless
    try:
        typer.echo(f"Cloning into '{project_name}'...")
        total = 900
        with typer.progressbar(length=total, label="Processing") as progress:
            demo_path = copy_from_template(progress, user_dict)
        typer.secho(
            f"Done it! Create Project {project_name} has success!\n"
            f"You can find it at: {demo_path}",
            fg=typer.colors.MAGENTA,
        )
    except Exception as e:
        typer.secho(
            f"Error!! create project {project_name} has error, errMsg: {e}",
            fg=typer.colors.MAGENTA,
            err=True,
        )


def create_mini():
    """
    Create mini project cli
    """
    user_dict = {}

    try:
        typer.echo(f"Cloning into ...")
        total = 900
        with typer.progressbar(length=total, label="Processing") as progress:
            demo_path = copy_from_template(progress, user_dict, os.path.normpath(os.getcwd()))
            compare_path = os.path.join(demo_path, "compareData")
            feat_files = get_files_from_dir(compare_path)

            config_ele = os.path.join(demo_path, "config/ele_locator.json")
            feat_files.append(config_ele)

            config_schema = os.path.join(demo_path, "config/schema_url.json")
            feat_files.append(config_schema)

            config_flybirds = os.path.join(demo_path, "config/flybirds_config.json")
            feat_files.append(config_flybirds)

            interface_path = os.path.join(demo_path, "interfaceIgnoreConfig/test.json")
            feat_files.append(interface_path)

            mock_path = os.path.join(demo_path, "mockCaseData/mock_test.json")
            feat_files.append(mock_path)

            for file in feat_files:
                os.remove(file)

            feature_path = os.path.join(demo_path, "features/test")
            shutil.rmtree(feature_path)

        typer.secho(
            f"Done it! Create Project has success!\n"
            f"You can find it at: {demo_path}",
            fg=typer.colors.MAGENTA,
        )
    except Exception as e:
        typer.secho(
            f"Error!! create project has error, errMsg: {e}",
            fg=typer.colors.MAGENTA,
            err=True,
        )


def copy_from_template(progress, user_dict, target=None):
    """
    Generate project files from template
    """
    # Serialization path
    src_file_path = template.__file__
    src_path = os.path.normpath(src_file_path[0: src_file_path.rfind(os.sep)])
    if target is None:
        target_path = os.path.normpath(
            os.path.join(os.path.normpath(os.getcwd()),
                         user_dict.get('project_name'))
        )
        if os.path.isdir(target_path):
            # target_path is existed
            shutil.rmtree(target_path)
    else:
        target_path = target

    shutil.copytree(src_path, target_path, dirs_exist_ok=True)
    progress.update(100)

    try:
        # process extend pkg
        add_extend_pkg(target_path)
    except Exception as e:
        log.error(f"[create_project][add_extend_pkg] has error, msg: {e}")
    progress.update(100)

    # delete file
    os.remove(os.path.normpath(os.path.join(target_path, "__init__.py")))
    progress.update(100)

    # modify platform
    test_platform = user_dict.get('test_platform')
    if test_platform is not None:
        replace_file_content(
            os.path.join(target_path, "config/flybirds_config.json"),
            "platform",
            test_platform,
        )
    progress.update(100)

    # modify deviceId
    device_id = user_dict.get('device_id')
    if device_id is not None:
        replace_file_content(
            os.path.join(target_path, "config/flybirds_config.json"),
            "deviceId",
            device_id,
        )
    progress.update(100)

    # modify packageName
    package_name = user_dict.get('package_name')
    if package_name is not None:
        replace_file_content(
            os.path.join(target_path, "config/flybirds_config.json"),
            "packageName",
            package_name,
        )
    progress.update(100)

    # modify webDriverAgent
    web_driver_agent = user_dict.get('web_driver_agent')
    if web_driver_agent is not None:
        replace_file_content(
            os.path.join(target_path, "config/flybirds_config.json"),
            "webDriverAgent",
            web_driver_agent,
        )
    progress.update(100)

    # modify browserType
    browser_type = user_dict.get('browser_type')
    if browser_type is not None:
        update_json_data(
            os.path.join(target_path, "config/flybirds_config.json"),
            "web_info.browserType",
            browser_type,
        )
    progress.update(100)

    # modify headless
    headless = user_dict.get('headless')
    if headless is not None:
        update_json_data(
            os.path.join(target_path, "config/flybirds_config.json"),
            "web_info.headless",
            headless,
        )
    progress.update(100)
    return target_path


def run_project(progress, target_path):
    """
    install packages
    """
    target_venv_path = os.path.normpath(
        os.path.join(target_path, activate_venv())
    )

    os.system(
        f"cd {target_path} && virtualenv venv && {target_venv_path} "
        f"&& pip install -r requirements.txt"
    )
    progress.update(100)


def activate_venv():
    """
    Activate venv
    :return:
    """
    sys_str = platform.system()
    if sys_str == "Windows":
        return "./venv/Scripts/activate"
    else:
        return "source venv/bin/activate"


def add_extend_pkg(demo_path):
    """
    Add expansion pack
    """
    pkg_query = "-flybirds-plugin"
    pkg_list = find_package_base_path(pkg_query)
    if pkg_list is None or len(pkg_list) <= 0:
        return

    copy_extend_files(pkg_list, demo_path)


def copy_extend_files(pkg_list, demo_pro_path):
    """
    Add extend features and config files
    """
    if demo_pro_path is None:
        return

    for pkg in pkg_list:
        # features src path
        extend_path = os.path.normpath(
            os.path.join(pkg.get("path"), pkg.get("name"), 'template'))

        if extend_path is None or not os.path.exists(extend_path):
            log.info(
                "[create_project][copy_extend_files]extend_path is none or not"
                "existed.")
            continue

        feat_path = os.path.join(os.path.normpath(extend_path), 'features')
        config_path = os.path.join(os.path.normpath(extend_path), 'config')
        custom_handle_path = os.path.join(os.path.normpath(extend_path),
                                          'pscript', "custom_handle")
        # add extend features
        if feat_path is not None and os.path.exists(feat_path):
            feat_files = get_files_from_dir(feat_path)

            # features target path
            demo_an_paths = get_paths_from_dir(demo_pro_path, 'android')
            demo_ios_paths = get_paths_from_dir(demo_pro_path, 'ios')

            for an_path in demo_an_paths:
                for file in feat_files:
                    shutil.copy(file, an_path)

            for ios_path in demo_ios_paths:
                for file in feat_files:
                    shutil.copy(file, ios_path)

        # add extend config
        if config_path is not None and os.path.exists(config_path):
            # config target path
            demo_config_path = os.path.join(os.path.normpath(demo_pro_path),
                                            'config')
            if os.path.isdir(demo_config_path):
                # target_path is existed
                shutil.rmtree(demo_config_path)
            shutil.copytree(config_path, demo_config_path)

        # add extend custom operation
        if custom_handle_path is not None and os.path.exists(
                custom_handle_path):
            # config target path
            demo_custom_handle_path = os.path.join(
                os.path.normpath(demo_pro_path),
                'pscript', "custom_handle")
            if os.path.isdir(demo_custom_handle_path):
                # target_path is existed
                shutil.rmtree(demo_custom_handle_path)
            shutil.copytree(custom_handle_path, demo_custom_handle_path)


def write_import_steps(pkg_list, demo_pro_path, site_path):
    """
    Write the steps that needs to be imported
    """
    if site_path is None:
        return

    import_str = ''
    # str that need to be imported
    for pkg in pkg_list:
        step_path = os.path.normpath(
            os.path.join(pkg.get("path"), pkg.get("name"), 'dsl', 'step'))

        if step_path is None or not os.path.exists(step_path):
            log.info(
                "[create_project][write_import_steps] extend_step path is none"
                "or not existed.")
            continue

        step_files = os.listdir(step_path)
        pkg_import_str = ''
        if step_files is not None and len(step_files) > 0:
            pkg_import_str = f'from {pkg.get("name")}.dsl.step import'
        for file in step_files:
            stem, suffix = os.path.splitext(file)
            if '__init__' == stem or '__pycache__' == stem:
                continue
            pkg_import_str += " " + stem + ","

        import_str += pkg_import_str.strip(',') + "\n"

    # write the extension steps that need to be imported into the file
    steps_file = os.path.join(os.path.normpath(demo_pro_path),
                              'features/steps/steps.py')
    update(steps_file, import_str)



# -*- coding: utf-8 -*-
"""
This module contains the App Core APIs.
"""

from flybirds.core.global_context import GlobalContext as g_context

"""
App Apis
"""


def wake_app(package_name, wait_time=None):
    """
    Start the target application on device
    :param package_name: package name
    :param wait_time:  wait time
    :return:
    """
    g_context.app.wake_app(package_name, wait_time)


def shut_app(package_name):
    """
    Stop the target application on device
    :param package_name: package name
    :return:
    """
    g_context.app.shut_app(package_name)


def install_app(package_path, wait_time=None):
    """
    Install application on device

    :param package_path: the path to file to be installed on target device
    :param wait_time:
    :return: None
    :platforms: Android
    :Example:
        >>> install_app(r"D:\\demo\\test.apk")
    """
    return g_context.app.install_app(package_path, wait_time)


def uninstall_app(package_name, wait_time=None):
    """
    Uninstall application on device

    :param package_name: name of the package, see also `start_app`
    :param wait_time:
    :return: None
    :platforms: Android
    :Example:
        >>> uninstall_app("com.flyBirds.music")
    """
    g_context.app.uninstall_app(package_name, wait_time)



# -*- coding: utf-8 -*-
"""
Poco element input
"""
from airtest.core.api import text, time

import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_ele as poco_ele
import flybirds.utils.flybirds_log as log
from flybirds.core.plugin.plugins.default.ui_driver.poco import poco_manage


def air_bdd_input(
        poco, select_dsl_str, optional, input_str, after_input_wait=None
):
    """
    input text
    :param poco:
    :param select_dsl_str:
    :param input_str:
    :param optional:
    :param after_input_wait:
    :return:
    """
    poco_ele.wait_exists(poco, select_dsl_str, optional)
    poco_object = poco_manage.create_poco_object_by_dsl(
        poco, select_dsl_str, optional
    )
    try:
        poco_object.set_text(input_str)
    except Exception as input_error:
        log.info(f"air_bdd_input has error:{str(input_error)}")
        poco_object.click()
        text(input_str)
    if not (after_input_wait is None):
        time.sleep(after_input_wait)



# -*- coding: utf-8 -*-
"""
Snapshots API
"""
import re

from flybirds.core.global_context import GlobalContext as g_context

__SOURCE__ = None
__IS_NEED_REFRESH__ = False


def clear_snap():
    """
    clear snapshot
    """
    global __SOURCE__
    __SOURCE__ = None


def get_snap():
    """
    Get snapshot
    """
    global __SOURCE__
    return __SOURCE__


def get_refresh_status():
    """
    Get if need refresh  snapshot status
    :return: 'False' or 'True'
    """
    global __IS_NEED_REFRESH__
    return __IS_NEED_REFRESH__


def fix_refresh_status(status):
    """
    Modify whether the snapshot status needs to be refreshed
    """
    global __IS_NEED_REFRESH__
    __IS_NEED_REFRESH__ = status


def refresh_snap():
    """
    Take screenshot
    """
    poco = g_context.element.ui_driver_init()
    frozen_poco = poco.freeze()
    global __SOURCE__
    __SOURCE__ = frozen_poco.agent.hierarchy.dump()


def snap_find(source, config):
    """
    find elements from snapshots
    text: text
    name: name, id
    textMatches: Regular
    """
    # print("config", config)
    text = config.get("text")
    name = config.get("name")
    textMatches = config.get("textMatches")
    elements = []

    def find(data):
        if data:
            if name:
                isMatch = data.get("name") == name
            elif text:
                isMatch = (
                    data.get("payload").get("text") or ""
                ).strip().replace(u"\u200b", "") == text.strip().replace(
                    u"\u200b", ""
                )
            elif textMatches:
                matchResult = re.search(
                    textMatches, data.get("payload").get("text") or ""
                )
                isMatch = not (matchResult is None)
            else:
                print("nothing to find")
            children = data.get("children") or []
            if isMatch:
                elements.append(data)
            elif children:
                nextEle = []
                if isinstance(
                    children,
                    (
                        frozenset,
                        list,
                        set,
                        tuple,
                    ),
                ):
                    nextEle = children
                else:
                    nextEle.append(children)
                for child in nextEle:
                    find(child)
            return elements
        else:
            print("not find data")

    return __SOURCE__ and find(__SOURCE__) or None
    # return source and find(source) or None



#! usr/bin/python
# -*- coding:utf-8 -*-
from .matchTemplate import MatchTemplate
from .sift import SIFT
from .utils import generate_result, get_keypoint_from_matches, keypoint_distance, rectangle_transform
from .exceptions import NoEnoughPointsError, PerspectiveTransformError, HomographyError, MatchResultError, InputImageError
from .base import BaseKeypoint



# -*- coding: utf-8 -*-
"""
dsl helper
"""
import base64
import re
from functools import wraps

import flybirds.core.global_resource as gr
import flybirds.utils.flybirds_log as log


# generate result_dic
def add_res_dic(dsl_params, functin_pattern, def_key):
    result_dic = {}
    match_obj = re.match(functin_pattern, dsl_params)
    if match_obj is not None:
        """
        senario：

        Flight, verifyEle=center_content_layout, verifyAction=position
        textMatches=shanghai.?
        .?economic.?, fuzzyMatch=true
        text=freshmode, timeout=15, swipeCount=40


        multi properities，example：text=freshmode, timeout=15, swipeCount=40
        Match from back to front, match back first,swipeCount=40
        match_obj_group_1（text=freshmode, timeout=15）
        f the conditions are still met, split again, Until the split to the
        last item: text=
        """
        group_1 = match_obj.group(1).strip().replace(u"\u200b", "")
        result_dic[match_obj.group(2)] = match_obj.group(3)
        match_obj_group_1 = re.match(functin_pattern, group_1)

        while match_obj_group_1 is not None:
            match_obj_group_1 = re.match(functin_pattern, group_1)
            if match_obj_group_1 is not None:
                group_1 = (
                    match_obj_group_1.group(1).strip().replace(u"\u200b", "")
                )
                result_dic[
                    match_obj_group_1.group(2)
                ] = match_obj_group_1.group(3)
            else:
                result_dic[def_key] = group_1
                break
        else:
            result_dic[def_key] = group_1

    else:
        result_dic[def_key] = dsl_params.strip().replace(u"\u200b", "")
    # print('result_dic44444', result_dic)
    return result_dic


# generate result_dic
def params_to_dic(dsl_params, def_key="selector"):
    """
    Convert the parameters in the dsl statement into dict format for use in
    subsequent processes
    """
    result_dic = {}
    functin_pattern = re.compile(r"([\S\s]+),\s*([a-zA-Z0-9_]+)\s*=\s*(\S+)")
    if isinstance(dsl_params, str):
        result_dic = add_res_dic(dsl_params, functin_pattern, def_key)
    log.info("result_dic: {}".format(result_dic))
    return result_dic


def split_must_param(dsl_params):
    """
    Get must and optional parameters
    """
    result = dsl_params.split(",", 1)
    result[0] = result[0].strip().replace(u"\u200b", "")
    result[1] = result[1].strip().replace(u"\u200b", "")
    return result


def get_params(context, *args):
    """
    Get param from context
    :param context: step context
    :param args: A tuple containing value and parameter name
    :return:
    """
    items = []
    for (val, param_name) in args:
        if val is not None:
            items.append(replace_str(val))
        elif hasattr(context, param_name):
            items.append(replace_str(getattr(context, param_name)))
    return items


def return_value(value, def_value=None):
    """
    get global attribute value
    """
    if value is not None:
        return value
    return def_value


def is_number(s):
    """
    Determine if the parameter is a number
    """
    try:
        float(s)
        return True
    except ValueError:
        log.error(f"param {s} is not a number!")
    try:
        import unicodedata
        unicodedata.numeric(s)
        return True
    except (TypeError, ValueError):
        log.error(f"param {s} cannot turn to a number!")
    return False


def replace_str(u_text):
    return u_text.strip().replace(u"\u200b", "")


def handle_str(un_handle_str):
    res = re.match(r"([\S\s]+),\s*[0-9_]+\s*", un_handle_str)
    if res is not None:
        return res.group(1)
    else:
        return un_handle_str


def str2bool(v):
    return v.lower() in ['true', '1', 't', 'y', 'yes', 'yeah', 'yup',
                         'certainly', 'uh-huh']


def get_use_define_param(context, param_name):
    use_define = context.get("use_define")
    log.info(f'use_define: {use_define}')
    params = [i for i in use_define if param_name + '=' in i]
    user_data = {}
    if len(params) > 0:
        if len(params) > 1:
            log.error(f'Cannot customize multiple parameters with the same '
                      f'name:{params}')
        value = params[0].split("=", 1)[1]
        user_data[param_name] = str(base64.b64decode(value), "utf-8")
    return user_data


def ele_wrap(func):
    @wraps(func)
    def wrapper_func(*args, **kwargs):
        context = args[0]
        for (k, v) in kwargs.items():
            if v is None:
                if hasattr(context, k):
                    v = getattr(context, k)
                else:
                    log.warn(f'[ele_wrap] step param:[{k}] is none.')
                    continue
            v = replace_str(v)
            if 'selector' in k:
                selector_str = v
                ele_key = v.split(',')[0]
                ele_value = gr.get_ele_locator(ele_key)
                v = selector_str.replace(ele_key, ele_value, 1)
            new_v = get_global_value(v)
            if new_v is not None:
                v = new_v
            kwargs[k] = v
        func(*args, **kwargs)
        # Do something after the function.

    return wrapper_func


def get_global_value(v):
    projectScript = gr.get_value("projectScript")
    if projectScript is not None:
        if hasattr(projectScript, "custom_operation"):
            custom_operation = projectScript.custom_operation
            if custom_operation is not None and hasattr(custom_operation,
                                                        "get_global_value"):
                rp = custom_operation.get_global_value(v)
                if rp is not None:
                    return rp
        elif hasattr(projectScript, "app_operation"):
            app_operation = projectScript.app_operation
            if app_operation is not None and hasattr(app_operation,
                                                     "get_global_value"):
                rp = app_operation.get_global_value(v)
                if rp is not None:
                    return rp
    return None



# -*- coding: utf-8 -*-
"""
This module defines the steps related to the UI element.
"""
from behave import step

from flybirds.core.global_context import GlobalContext as g_Context
from flybirds.utils.dsl_helper import ele_wrap


@step("text[{selector}]property[{param2}]is {param3}")
@ele_wrap
def text_attr_equal(context, selector=None, param2=None, param3=None):
    """
    Check if the value of the attribute param2 of the text element param1 in
     the page is param3

    :param context: step context
    :param selector: locator string for text element (or None).
    :param param2: attribute Name
    :param param3: expected Value
    """
    g_Context.step.text_attr_equal(context, selector, param2, param3)


@step("element[{selector}]property[{param2}]is {param3}")
@ele_wrap
def ele_attr_equal(context, selector=None, param2=None, param3=None):
    """
    Check if the value of the attribute param2 of the selector element param1
     in the page is param3

    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2: attribute Name
    :param param3: expected Value
    """
    g_Context.step.ele_attr_equal(context, selector, param2, param3)


@step("click[{selector}]")
@ele_wrap
def click_ele(context, selector=None):
    """
    Click on the selector element
    :param context: step context
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.click_ele(context, selector)


@step("click text[{selector}]")
@ele_wrap
def click_text(context, selector=None):
    """
    Click on the text element
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.click_text(context, selector)


@step("click ocr text[{selector}]")
@ele_wrap
def click_ocr_text(context, selector=None):
    """
    Click on the ocr text element
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.click_ocr_text(context, selector)


@step("click image[{selector}]")
@ele_wrap
def click_image(context, selector=None):
    """
    Click on the image
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.click_image(context, selector)


@step("click position[{x},{y}]")
@ele_wrap
def click_coordinates(context, x=None, y=None):
    """
    Click on the screen coordinates
    :param context: step context
    :param x: Coordinate x-axis
    :param y: Coordinate y-axis.
    """
    g_Context.step.click_coordinates(context, x, y)


@step("in[{selector}]input[{param2}]")
@ele_wrap
def ele_input(context, selector=None, param2=None):
    """
    Enter the value param2 in the selector element param1
    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2: string to be input
    """
    g_Context.step.ele_input(context, selector, param2)


@step("in ocr[{selector}]input[{param2}]")
@ele_wrap
def ocr_text_input(context, selector=None, param2=None):
    """
    Enter the value param2 in the selector element param1
    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2: string to be input
    """
    g_Context.step.ocr_text_input(context, selector, param2)


@step("element[{selector}]position not change in[{param2}]seconds")
@ele_wrap
def position_not_change(context, selector=None, param2=None):
    """
    Check that the position of the selector element param1 has not changed
    within param2 seconds
    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2:
    """
    g_Context.step.position_not_change(context, selector, param2)


@step("[{selector}]slide to {param2} distance[{param3}]")
@ele_wrap
def ele_swipe(context, selector=None, param2=None, param3=None):
    """
    Selector element param1 slides in the specified direction param2 and
    slides the specified distance param3
    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2: slide direction (top/bottom/left/right)
    :param param3: slide distance
    """
    g_Context.step.ele_swipe(context, selector, param2, param3)


@step("slide to {param1} distance[{param2}]")
@ele_wrap
def full_screen_swipe(context, param1=None, param2=None):
    """
    Slide the full screen in the specified direction for the specified distance
    :param context: step context
    :param param1: slide direction (top/bottom/left/right)
    :param param2: slide distance
    """
    g_Context.step.full_screen_swipe(context, param1, param2)


@step("exist text[{selector}]")
@ele_wrap
def wait_text_exist(context, selector=None):
    """
    The specified text element string exists in the page
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.wait_text_exist(context, selector)


@step("ocr exist text[{selector}]")
@ele_wrap
def ocr_text_exist(context, selector=None):
    """
    The specified text element string exists in the page
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.ocr_text_exist(context, selector)


@step("ocr contain text[{selector}]")
@ele_wrap
def ocr_text_contain(context, selector=None):
    """
    The specified text element string exists in the page
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.ocr_text_contain(context, selector)


@step("page ocr complete find text[{selector}]")
@ele_wrap
def wait_ocr_text_appear(context, selector=None):
    """
    Wait for the page to finish rendering and the selector element param1
     to appear
    :param context: step context
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.wait_ocr_text_appear(context, selector)


@step("not exist text[{selector}]")
@ele_wrap
def text_not_exist(context, selector=None):
    """
    The specified text element string does not exist in the page
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.text_not_exist(context, selector)


@step("ocr not exist text[{selector}]")
@ele_wrap
def ocr_text_not_exist(context, selector=None):
    """
    The specified text string does not exist in the page
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.ocr_text_not_exist(context, selector)

@step("text[{selector}]disappear")
@ele_wrap
def wait_text_disappear(context, selector=None):
    """
    The specified text element string disappears from the page within
     a specified period of time
    :param context: step context
    :param selector: locator string for text element (or None).
    """
    g_Context.step.wait_text_disappear(context, selector)


@step("exist [{p_selector}] subNode [{c_selector}] element")
@ele_wrap
def find_child_from_parent(context, p_selector=None, c_selector=None):
    """
    The specified child selector element of the specified parent selector
    element exists in the page.
    :param context: step context
    :param p_selector: locator string for parent selector element (or None).
    :param c_selector: locator string for selector child element (or None).
    """
    g_Context.step.find_child_from_parent(context, p_selector, c_selector)


@step("exist[{selector}]element")
@ele_wrap
def wait_ele_exit(context, selector=None):
    """
    The specified selector element string exists in the page
    :param context: step context
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.wait_ele_exit(context, selector)


@step("not exist element[{selector}]")
@ele_wrap
def ele_not_exit(context, selector=None):
    """
    The specified selector element string does not exists in the page
    :param context: step context
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.ele_not_exit(context, selector)


@step("element[{selector}]disappear")
@ele_wrap
def wait_ele_disappear(context, selector=None):
    """
    The specified selector element string disappears from the page within
     a specified period of time
    :param context: step context
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.wait_ele_disappear(context, selector)


@step("the text of element[{selector}]is[{param2}]")
@ele_wrap
def ele_text_equal(context, selector=None, param2=None):
    """
    Check if the value of the text of the selector element param1 is param2
    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2: expected value
    """
    g_Context.step.ele_text_equal(context, selector, param2)


@step("the text of element[{selector}]include[{param2}]")
@ele_wrap
def ele_text_container(context, selector=None, param2=None):
    """
   Check if the value of the text of the selector element param1 include param2
    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2: expected value
    """
    g_Context.step.ele_text_container(context, selector, param2)


@step("page rendering complete appears element[{selector}]")
@ele_wrap
def wait_ele_appear(context, selector=None):
    """
    Wait for the page to finish rendering and the selector element param1
     to appear
    :param context: step context
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.wait_ele_appear(context, selector)


@step("existing element[{selector}]")
@ele_wrap
def exist_ele(context, selector=None):
    """
    The specified selector element string exists in the page
    :param context: step context
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.exist_ele(context, selector)


@step("in[{p_selector}]from {param2} find[{c_selector}]element")
@ele_wrap
def swipe_to_ele(context, p_selector=None, param2=None, c_selector=None):
    """
    Within the specified selector element Slide in the specified direction
     to find the selector element
    :param context: step context
    :param p_selector: locator string for parent selector element (or None).
    :param param2: slide direction (top/bottom/left/right)
    :param c_selector: locator string for selector child element (or None).
    """
    g_Context.step.swipe_to_ele(context, p_selector, param2, c_selector)


@step("from {param1} find[{selector}]element")
@ele_wrap
def full_screen_swipe_to_ele_aaa(context, param1=None, selector=None):
    """
    Full screen swipe in the specified direction to find the specified
     selector element
     :param context: step context
    :param param1: slide direction (top/bottom/left/right)
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.full_screen_swipe_to_ele_aaa(context, param1, selector)


@step("from {param1} find[{selector}]text")
@ele_wrap
def full_screen_swipe_to_ocr_txt(context, param1=None, selector=None):
    """
    Full screen swipe in the specified direction to find the specified
     selector element
     :param context: step context
    :param param1: slide direction (top/bottom/left/right)
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.full_screen_swipe_to_ocr_txt(context, param1, selector)


@step("from {param1} find[{selector}]image")
@ele_wrap
def full_screen_swipe_to_img(context, param1=None, selector=None):
    """
    Full screen swipe in the specified direction to find the specified
     selector element
     :param context: step context
    :param param1: slide direction (top/bottom/left/right)
    :param selector: locator string for selector element (or None).
    """
    g_Context.step.full_screen_swipe_to_img(context, param1, selector)


@step("clear [{selector}] and input[{param2}]")
@ele_wrap
def ele_clear_input(context, selector=None, param2=None):
    """
    Empty the selector element param1 and enter the value param2
    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2: string to be input
    """
    g_Context.step.ele_clear_input(context, selector, param2)


@step("from [{selector}] select [{param2}]")
@ele_wrap
def ele_select(context, selector=None, param2=None):
    """
    Select the value param2 from the dropdown box element param1
    :param context: step context
    :param selector: locator string for selector element (or None).
    :param param2: text or value of select option
    """
    g_Context.step.ele_select(context, selector, param2)


@step(
    "the text of element [{p_selector}] subNode [{c_selector}] is [{param3}]")
@ele_wrap
def find_text_from_parent(context, p_selector=None, c_selector=None,
                          param3=None):
    """
    check the text of the child element in the parent element is param3
    :param context: step context
    :param p_selector: locator string for parent selector element (or None).
    :param c_selector: locator string for selector child element (or None).
    :param param3: expected value.
    """
    g_Context.step.find_text_from_parent(context, p_selector, c_selector,
                                         param3)

@step("exist image [{param}]")
def img_exist(context, param):
    g_Context.step.img_exist(context, param)


@step("not exist image [{param}]")
def img_not_exist(context, param):
    g_Context.step.img_not_exist(context, param)




# -*- coding: utf-8 -*-
"""
step extend manage
"""
import os

from flybirds.utils.pkg_helper import load_pkg_by_ns
from flybirds.utils.flybirds_log import logger


def load_steps():
    try:
        if os.environ.get("extend_pkg_list") is not None:
            extend_pkg = os.environ.get("extend_pkg_list")
            extend_pkg_list = extend_pkg.split(",")
            if len(extend_pkg_list) > 0:
                for pkg in extend_pkg_list:
                    if pkg is not None and pkg != "":
                        logger.info(f"load extend package:{pkg}")
                        load_pkg_by_ns(f"{pkg}.dsl.step")
    except Exception as load_ex:
        logger.info(f"load extend package error :f{load_ex}")



# -*- coding: utf-8 -*-
"""
snippet helper
"""
import os
import subprocess
import re
import flybirds.core.driver.device as device


def create_sub_process(cmd):
    """
    Create a child process that executes a specific command
    """
    proc = subprocess.Popen(
        cmd,
        cwd=os.getcwd(),
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        creationflags=0,
    )
    return proc


def list_comparator(o_list, t_list):
    """
    Determine whether two lists contain the same elements
    :param o_list:
    :param t_list:
    :return:
    """

    try:
        if sorted(o_list) == sorted(t_list):
            return True
    except Exception:
        return False
    else:
        return False


def chose_first_number(string):
    """
    Take out the first number in the statement
    :param string:
    :return:
    """
    return int(re.search(r"\d+", string).group(0))


def schema_goto(
    page_name,
    schema_url,
    schema_goto_module,
    deal_method=None,
    params_deal_module=None,
):
    """
    schema go to
    """
    if not (deal_method is None):
        deal_method = getattr(params_deal_module, deal_method)
        schema_url = deal_method(schema_url)

    schema_rule = getattr(schema_goto_module, "schema_deal_rule")
    schema_url = schema_rule(page_name, schema_url)
    cmd = "am start -a 'android.intent.action.VIEW' -d '" + schema_url + "'"
    device.use_shell(cmd)



# -*- coding: utf-8 -*-
"""
Overriding poco snapshot
"""
import base64
import os
import time

from airtest import aircv
from airtest.core.helper import G
from airtest.core.settings import Settings as ST

import flybirds.core.global_resource as gr


def get_screen():
    save_path = snapshot()
    return base64.b64encode(open(save_path, 'rb').read()), 'png'


def snapshot(filename=None, msg="", quality=None, max_size=None):
    """
    Take the screenshot of the target device and save it to the file.
    """
    if not quality:
        quality = ST.SNAPSHOT_QUALITY
    if not max_size and ST.IMAGE_MAXSIZE:
        max_size = ST.IMAGE_MAXSIZE
    if filename:
        if not os.path.isabs(filename):
            log_dir = ST.LOG_DIR or "."
            filename = os.path.join(log_dir, filename)
        screen = G.DEVICE.snapshot(filename, quality=quality,
                                   max_size=max_size)
        return try_log_screen(screen, quality=quality, max_size=max_size)
    else:
        return try_log_screen(quality=quality, max_size=max_size)


def try_log_screen(screen=None, quality=None, max_size=None):
    """
    Save screenshot to file

    Args:
        screen: screenshot to be saved
        quality: The image quality, default is ST.SNAPSHOT_QUALITY
        max_size: the maximum size of the picture, e.g 1200

    Returns:
       filepath
    """
    if not ST.LOG_DIR:
        ST.LOG_DIR = gr.get_screen_save_dir()
    if not ST.SAVE_IMAGE:
        return
    if not quality:
        quality = ST.SNAPSHOT_QUALITY
    if not max_size:
        max_size = ST.IMAGE_MAXSIZE
    if screen is None:
        screen = G.DEVICE.snapshot(quality=quality)
    filename = "%(time)d.jpg" % {'time': time.time() * 1000}
    filepath = os.path.join(ST.LOG_DIR, filename)
    if screen is not None:
        aircv.imwrite(filepath, screen, quality, max_size=max_size)
        return filepath
    return None



# -*- coding: utf-8 -*-
"""
language help
"""
from behave.i18n import languages

from flybirds.core.dsl.globalization.i18n import globalization
from flybirds.core.dsl.globalization.i18n import step_language


def parse_keyword(text, language):
    """
    parse bh key word
    """
    kw = language
    for kw in languages[language][text]:
        if kw.endswith("<"):
            kw = kw[:-1]
    return kw


def parse_glb_str(text, language):
    """
    parse glb str
    """
    if globalization.__contains__(language):
        sn = globalization[language]
        if sn.__contains__(text):
            return sn[text]
    return None


def parse_glb_step(text, language):
    """
    parse glb step
    """
    if step_language.__contains__(language):
        sn = step_language[language]
        if sn.__contains__(text):
            return sn[text]
    return None


def get_language_list():
    """
    get i18 global step language list
    """
    lg_list = []
    if step_language is not None:
        for name in step_language.keys():
            lg_list.append(name)

    return lg_list


def get_glb_key(kw, language):
    """
    get glb dict key by keyword
    """
    if globalization.__contains__(language):
        l_dict = globalization[language]
        if l_dict is not None:
            for k, v in l_dict.items():
                if v == kw:
                    return k
    return None



# -*- coding: utf-8 -*-
"""
Element swipe
"""
import flybirds.core.global_resource as gr
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_swipe as ps
import flybirds.utils.dsl_helper as dsl_helper
import flybirds.utils.point_helper as point_helper


def ele_swipe(context, param1, param2, param3):
    poco_instance = gr.get_value("pocoInstance")

    param1_dict = dsl_helper.params_to_dic(param1)
    selector_str = param1_dict["selector"]
    optional = {}
    if "path" in param1_dict.keys():
        optional["path"] = param1_dict["path"]
    elif "multiSelector" in param1_dict.keys():
        optional["multiSelector"] = param1_dict["multiSelector"]
    if "timeout" in param1_dict.keys():
        optional["timeout"] = float(param1_dict["timeout"])
    else:
        optional["timeout"] = gr.get_frame_config_value("wait_ele_timeout", 10)

    param3_dict = dsl_helper.params_to_dic(param3, "swipeNumber")

    start_point = [0.5, 0.5]
    if "startX" in param3_dict.keys():
        start_point[0] = float(param3_dict["startX"])
    if "startY" in param3_dict.keys():
        start_point[1] = float(param3_dict["startY"])

    screen_size = gr.get_device_size()

    direction = point_helper.search_direction_switch(param2.strip())

    distance = float(param3_dict["swipeNumber"])

    duration = None
    if gr.get_frame_config_value("use_swipe_duration", False):
        duration = gr.get_frame_config_value("swipe_duration", 1)
    if "duration" in param3_dict.keys():
        duration = float(param3_dict["duration"])

    ready_time = gr.get_frame_config_value("swipe_ready_time", 1)
    if "readyTime" in param3_dict.keys():
        ready_time = float(param3_dict["readyTime"])

    ps.air_bdd_ele_swipe(
        poco_instance,
        selector_str,
        optional,
        start_point,
        screen_size,
        direction,
        distance,
        duration,
        ready_time,
    )


def full_screen_swipe(context, param1, param2):
    poco_instance = gr.get_value("pocoInstance")

    param2_dict = dsl_helper.params_to_dic(param2, "swipeNumber")

    start_point = [0.5, 0.5]
    if "startX" in param2_dict.keys():
        start_point[0] = float(param2_dict["startX"])
    if "startY" in param2_dict.keys():
        start_point[1] = float(param2_dict["startY"])

    screen_size = gr.get_device_size()

    direction = point_helper.search_direction_switch(param1.strip())

    distance = float(param2_dict["swipeNumber"])

    duration = None
    if gr.get_frame_config_value("use_swipe_duration", False):
        duration = gr.get_frame_config_value("swipe_duration", 1)
    if "duration" in param2_dict.keys():
        duration = float(param2_dict["duration"])

    ready_time = gr.get_frame_config_value("swipe_ready_time", 1)
    if "readyTime" in param2_dict.keys():
        ready_time = float(param2_dict["readyTime"])

    ps.air_bdd_full_screen_swipe(
        poco_instance,
        start_point,
        screen_size,
        direction,
        distance,
        duration,
        ready_time,
    )



#! usr/bin/python
# -*- coding:utf-8 -*-
import cv2
import numpy as np
from baseImage import Image, Rect

from .matchTemplate import MatchTemplate
from .utils import (generate_result, get_keypoint_from_matches, keypoint_distance, rectangle_transform)
from .exceptions import (NoEnoughPointsError, PerspectiveTransformError, HomographyError, MatchResultError,
                         InputImageError)
from typing import List


class BaseKeypoint(object):
    FILTER_RATIO = 1
    METHOD_NAME = None
    Dtype = None
    Place = None
    template = MatchTemplate()

    def __init__(self, threshold=0.8, rgb=True, **kwargs):
        """
        init

        Args:
            threshold: recognition threshold(0~1)
            rgb: Whether to use rgb channel for verification
        """
        self.threshold = threshold
        self.rgb = rgb
        self.detector = self.create_detector(**kwargs)
        self.matcher = self.create_matcher(**kwargs)

    def create_matcher(self, **kwargs):
        raise NotImplementedError

    def create_detector(self, **kwargs):
        raise NotImplementedError

    def find_best_result(self, im_source, im_search, threshold=None, rgb=None, **kwargs):
        """
        Through feature point matching, find the range that best matches im_search in im_source

        Args:
            im_source: image to be matched
            im_search: image template
            threshold: recognition threshold(0~1)
            rgb: Whether to use rgb channel for verification

        Returns:

        """
        max_count = 1
        ret = self.find_all_results(im_source=im_source, im_search=im_search, threshold=threshold, rgb=rgb,
                                    max_count=max_count, **kwargs)
        if ret:
            return ret[0]
        return None

    def find_all_results(self, im_source, im_search, threshold=None, rgb=None, max_count=10, max_iter_counts=20,
                         distance_threshold=150):
        """
        Through feature point matching, find all the ranges that match im_search in im_source

        Args:
             im_source: the image to be matched
             im_search: image template
             threshold: recognition threshold (0~1)
             rgb: whether to use the rgb channel for verification
             max_count: the maximum number of matches that can be returned
             max_iter_counts: The maximum number of searches, which needs to be greater than max_count
             distance_threshold: distance threshold, after the feature point (first_point) is greater than
                                 the threshold, no subsequent screening will be done

        Returns:

        """
        threshold = self.threshold if threshold is None else threshold
        rgb = self.rgb if rgb is None else rgb

        im_source, im_search = self.input_image_check(im_source, im_search)
        result = []
        if im_source.channels == 1:
            rgb = False

        kp_src, des_src = self.get_keypoint_and_descriptor(image=im_source)
        kp_sch, des_sch = self.get_keypoint_and_descriptor(image=im_search)

        kp_src, kp_sch = list(kp_src), list(kp_sch)
        # In the feature point set, match the closest feature point
        matches = np.array(self.match_keypoint(des_sch=des_sch, des_src=des_src))
        kp_sch_point = np.array([(kp.pt[0], kp.pt[1], kp.angle) for kp in kp_sch])
        kp_src_matches_point = np.array([[(*kp_src[dMatch.trainIdx].pt, kp_src[dMatch.trainIdx].angle)
                                          if dMatch else np.nan for dMatch in match] for match in matches])
        _max_iter_counts = 0
        src_pop_list = []
        while True:
            # not use of matches to judge nan, because the type is wrong
            if (np.count_nonzero(~np.isnan(kp_src_matches_point)) == 0) or (len(result) == max_count) or (
                    _max_iter_counts >= max_iter_counts):
                break
            _max_iter_counts += 1
            filtered_good_point, angle, first_point = self.filter_good_point(matches=matches, kp_src=kp_src,
                                                                             kp_sch=kp_sch,
                                                                             kp_sch_point=kp_sch_point,
                                                                             kp_src_matches_point=kp_src_matches_point)
            if first_point.distance > distance_threshold:
                break

            rect, confidence = None, 0
            try:
                rect, confidence = self.extract_good_points(im_source=im_source, im_search=im_search, kp_src=kp_src,
                                                            kp_sch=kp_sch, good=filtered_good_point, angle=angle,
                                                            rgb=rgb)
                # print(f'good:{len(filtered_good_point)}, rect={rect}, confidence={confidence}')
            except PerspectiveTransformError:
                pass
            finally:

                if rect and confidence >= threshold:
                    br, tl = rect.br, rect.tl
                    for index, match in enumerate(kp_src_matches_point):
                        x, y = match[:, 0], match[:, 1]
                        flag = np.argwhere((x < br.x) & (x > tl.x) & (y < br.y) & (y > tl.y))
                        for _index in flag:
                            src_pop_list.append(matches[index, _index][0].trainIdx)
                            kp_src_matches_point[index, _index, :] = np.nan
                            matches[index, _index] = np.nan
                    result.append(generate_result(rect, confidence))
                else:
                    for match in filtered_good_point:
                        flags = np.argwhere(matches[match.queryIdx, :] == match)
                        for _index in flags:
                            kp_src_matches_point[match.queryIdx, _index, :] = np.nan
                            matches[match.queryIdx, _index] = np.nan
        return result

    def get_keypoint_and_descriptor(self, image):
        """
        Get image keypoint (keypoint) and descriptor (descriptor)

        Args:
            image: Grayscale image to be detected

        Returns:

        """
        if image.channels == 3:
            image = image.cvtColor(cv2.COLOR_BGR2GRAY).data
        else:
            image = image.data
        keypoint, descriptor = self.detector.detectAndCompute(image, None)

        # if len(keypoint) < 2:
        #    raise NoEnoughPointsError('{} detect not enough feature points in input images'.format(self.METHOD_NAME))
        return keypoint, descriptor

    @staticmethod
    def filter_good_point(matches, kp_src, kp_sch, kp_sch_point, kp_src_matches_point):
        """ Filter the sweet spot """
        # Assume that the first point and the point with the smallest distance are the reference points
        sort_list = [sorted(match, key=lambda x: x is np.nan and float('inf') or x.distance)[0]
                     for match in matches]
        sort_list = [v for v in sort_list if v is not np.nan]

        first_good_point: cv2.DMatch = sorted(sort_list, key=lambda x: x.distance)[0]
        first_good_point_train: cv2.KeyPoint = kp_src[first_good_point.trainIdx]
        first_good_point_query: cv2.KeyPoint = kp_sch[first_good_point.queryIdx]
        first_good_point_angle = first_good_point_train.angle - first_good_point_query.angle

        def get_points_origin_angle(point_x, point_y, offset):
            points_origin_angle = np.arctan2(
                (point_y - offset.pt[1]),
                (point_x - offset.pt[0])
            ) * 180 / np.pi

            points_origin_angle = np.where(
                points_origin_angle == 0,
                points_origin_angle, points_origin_angle - offset.angle
            )
            points_origin_angle = np.where(
                points_origin_angle >= 0,
                points_origin_angle, points_origin_angle + 360
            )
            return points_origin_angle

        # Calculate the rotation angle between this point and other feature points on the template image
        first_good_point_sch_origin_angle = get_points_origin_angle(kp_sch_point[:, 0], kp_sch_point[:, 1],
                                                                    first_good_point_query)

        # Calculate the angle between the point and other feature points in the target image
        kp_sch_rotate_angle = kp_sch_point[:, 2] + first_good_point_angle
        kp_sch_rotate_angle = np.where(kp_sch_rotate_angle >= 360, kp_sch_rotate_angle - 360, kp_sch_rotate_angle)
        kp_sch_rotate_angle = kp_sch_rotate_angle.reshape(kp_sch_rotate_angle.shape + (1,))

        kp_src_angle = kp_src_matches_point[:, :, 2]
        good_point = np.array([matches[index][array[0]] for index, array in
                               enumerate(np.argsort(np.abs(kp_src_angle - kp_sch_rotate_angle)))])

        # Calculate the rotation angle of each point with first_good_point as the origin
        good_point_nan = (np.nan, np.nan)
        good_point_pt = np.array([good_point_nan if dMatch is np.nan else (*kp_src[dMatch.trainIdx].pt,)
                                  for dMatch in good_point])
        good_point_origin_angle = get_points_origin_angle(good_point_pt[:, 0], good_point_pt[:, 1],
                                                          first_good_point_train)
        threshold = round(5 / 360, 2) * 100
        point_bool = (np.abs(good_point_origin_angle - first_good_point_sch_origin_angle) / 360) * 100 < threshold
        _, index = np.unique(good_point_pt[point_bool], return_index=True, axis=0)
        good = good_point[point_bool]
        good = good[index]
        return good, int(first_good_point_angle), first_good_point

    def match_keypoint(self, des_sch, des_src, k=10):
        """
        Feature point matching

        Args:
             des_src: descriptor set of the image to be matched
             des_sch: descriptor set for image templates
             k(int): how many matching points to get

        Returns:
            List[List[cv2.DMatch]]: contains the best matching descriptor
        """
        # k=2 means that each feature point takes out the 2 most matching corresponding points
        matches = self.matcher.knnMatch(des_sch, des_src, k)
        return matches

    def get_good_in_matches(self, matches):
        """
        Feature point filtering

        Args:
            matches: Feature point set

        Returns:
            List[cv2.DMatch]: Filtered descriptor set
        """
        if not matches:
            return None
        good = []
        for match_index in range(len(matches)):
            match = matches[match_index]
            for DMatch_index in range(len(match)):
                if match[DMatch_index].distance <= self.FILTER_RATIO * match[-1].distance:
                    good.append(match[DMatch_index])
        return good

    def extract_good_points(self, im_source, im_search, kp_src, kp_sch, good, angle, rgb):
        """
        According to the number of matching points (good), extract the recognition area

        Args:
             im_source: the image to be matched
             im_search: image template
             kp_src: keypoint set
             kp_sch: keypoint set
             good: descriptor set
             angle: rotation angle
             rgb: whether to use the rgb channel for verification

        Returns:
            range, and confidence
        """
        len_good = len(good)
        confidence, rect, target_img = None, None, None

        if len_good == 0:
            pass
        elif len_good == 1:
            target_img, rect = self._handle_one_good_points(im_source=im_source, im_search=im_search,
                                                            kp_sch=kp_sch, kp_src=kp_src, good=good, angle=angle)
        elif len_good == 2:
            target_img, rect = self._handle_two_good_points(im_source=im_source, im_search=im_search,
                                                            kp_sch=kp_sch, kp_src=kp_src, good=good, angle=angle)
        elif len_good == 3:
            target_img, rect = self._handle_three_good_points(im_source=im_source, im_search=im_search,
                                                              kp_sch=kp_sch, kp_src=kp_src, good=good, angle=angle)
        else:  # len > 4
            target_img, rect = self._handle_many_good_points(im_source=im_source, im_search=im_search,
                                                             kp_sch=kp_sch, kp_src=kp_src, good=good)

        if target_img:
            confidence = self._cal_confidence(im_source=im_search, im_search=target_img, rgb=rgb)

        return rect, confidence

    def _handle_one_good_points(self, im_source, im_search, kp_src, kp_sch, good, angle):
        """
        When the number of feature point matching is equal to 1, the rectangle is scaled according
        to the size of the feature point, and the recognized target image is obtained according to
        the rotation angle.

        Args:
             im_source: the image to be matched
             im_search: image template
             kp_sch: keypoint set
             kp_src: keypoint set
             good: descriptor set
             angle: rotation angle

        Returns:
            Image to be verified
        """
        sch_point = get_keypoint_from_matches(kp=kp_sch, matches=good, mode='query')[0]
        src_point = get_keypoint_from_matches(kp=kp_src, matches=good, mode='train')[0]

        scale = src_point.size / sch_point.size
        h, w = im_search.size
        _h, _w = h * scale, w * scale
        src = np.float32(rectangle_transform(point=sch_point.pt, size=(h, w), mapping_point=src_point.pt,
                                             mapping_size=(_h, _w), angle=angle))
        dst = np.float32([[0, 0], [w, 0], [0, h], [w, h]])
        output = self._perspective_transform(im_source=im_source, im_search=im_search, src=src, dst=dst)
        rect = self._get_perspective_area_rect(im_source=im_source, src=src)
        return output, rect

    def _handle_two_good_points(self, im_source, im_search, kp_src, kp_sch, good, angle):
        """
        When the number of feature points matching is equal to 2, the rectangle is scaled
        according to the distance difference between the two points, and the recognized target
        image is obtained according to the rotation angle.

        Args:
             im_source: the image to be matched
             im_search: image template
             kp_sch: keypoint set
             kp_src: keypoint set
             good: descriptor set
             angle: rotation angle

        Returns:
            Image to be verified
        """
        sch_point = get_keypoint_from_matches(kp=kp_sch, matches=good, mode='query')
        src_point = get_keypoint_from_matches(kp=kp_src, matches=good, mode='train')

        sch_distance = keypoint_distance(sch_point[0], sch_point[1])
        src_distance = keypoint_distance(src_point[0], src_point[1])

        try:
            scale = src_distance / sch_distance  # Calculate the zoom size
        except ZeroDivisionError:
            if src_distance == sch_distance:
                scale = 1
            else:
                return None, None

        h, w = im_search.size
        _h, _w = h * scale, w * scale
        src = np.float32(rectangle_transform(point=sch_point[0].pt, size=(h, w), mapping_point=src_point[0].pt,
                                             mapping_size=(_h, _w), angle=angle))
        dst = np.float32([[0, 0], [w, 0], [0, h], [w, h]])
        output = self._perspective_transform(im_source=im_source, im_search=im_search, src=src, dst=dst)
        rect = self._get_perspective_area_rect(im_source=im_source, src=src)
        return output, rect

    def _handle_three_good_points(self, im_source, im_search, kp_src, kp_sch, good, angle):
        """
        When the number of feature points matching is equal to 3, the rectangle is scaled
        according to the difference in the area of the triangle formed by the three points,
        and the recognized target image is obtained according to the rotation angle.

        Args:
             im_source: the image to be matched
             im_search: image template
             kp_sch: keypoint set
             kp_src: keypoint set
             good: descriptor set
             angle: rotation angle

        Returns:
            Image to be verified
        """
        sch_point = get_keypoint_from_matches(kp=kp_sch, matches=good, mode='query')
        src_point = get_keypoint_from_matches(kp=kp_src, matches=good, mode='train')

        def _area(point_list):
            p1_2 = keypoint_distance(point_list[0], point_list[1])
            p1_3 = keypoint_distance(point_list[0], point_list[2])
            p2_3 = keypoint_distance(point_list[1], point_list[2])

            s = (p1_2 + p1_3 + p2_3) / 2
            area = (s * (s - p1_2) * (s - p1_3) * (s - p2_3)) ** 0.5
            return area

        sch_area = _area(sch_point)
        src_area = _area(src_point)

        try:
            scale = src_area / sch_area  # Calculate the zoom size
        except ZeroDivisionError:
            if sch_area == src_area:
                scale = 1
            else:
                return None, None

        h, w = im_search.size
        _h, _w = h * scale, w * scale
        src = np.float32(rectangle_transform(point=sch_point[0].pt, size=(h, w), mapping_point=src_point[0].pt,
                                             mapping_size=(_h, _w), angle=angle))
        dst = np.float32([[0, 0], [w, 0], [0, h], [w, h]])
        output = self._perspective_transform(im_source=im_source, im_search=im_search, src=src, dst=dst)
        rect = self._get_perspective_area_rect(im_source=im_source, src=src)
        return output, rect

    def _handle_many_good_points(self, im_source, im_search, kp_src, kp_sch, good):
        """
        When the number of feature point matching is >= 4,
        use single matrix mapping to obtain the recognized target image

        Args:
             im_source: the image to be matched
             im_search: image template
             kp_sch: keypoint set
             kp_src: keypoint set
             good: descriptor set

        Returns:
            Perspective transformed image
        """

        sch_pts, img_pts = np.float32([kp_sch[m.queryIdx].pt for m in good]).reshape(
            -1, 1, 2), np.float32([kp_src[m.trainIdx].pt for m in good]).reshape(-1, 1, 2)
        # M is the transformation matrix
        M, mask = self._find_homography(sch_pts, img_pts)
        # Calculate the transformed coordinates of the four corner matrices,
        # that is, the vertex coordinates of the target area in the large image:
        h, w = im_search.size
        h_s, w_s = im_source.size
        pts = np.float32([[0, 0], [0, h - 1], [w - 1, h - 1], [w - 1, 0]]).reshape(-1, 1, 2)
        try:
            dst: np.ndarray = cv2.perspectiveTransform(pts, M)
            # img = im_source.clone().data
            # img2 = cv2.polylines(img, [np.int32(dst)], True, 255, 3, cv2.LINE_AA)
            # Image(img).imshow('dst')
            pypts = [tuple(npt[0]) for npt in dst.tolist()]
            src = np.array([pypts[0], pypts[3], pypts[1], pypts[2]], dtype=np.float32)
            dst = np.float32([[0, 0], [w, 0], [0, h], [w, h]])
            output = self._perspective_transform(im_source=im_source, im_search=im_search, src=src, dst=dst)
        except cv2.error as err:
            raise PerspectiveTransformError(err)

        # img = im_source.clone().data
        # cv2.polylines(img, [np.int32(dst)], True, 255, 3, cv2.LINE_AA)
        # Image(img).imshow()
        # cv2.waitKey(0)

        rect = self._get_perspective_area_rect(im_source=im_source, src=src)
        return output, rect

    @staticmethod
    def _target_image_crop(img, rect):
        """
        Capture target image

        Args:
            img: image
            rect: Image range

        Returns:
            cropped image
        """
        try:
            target_img = img.crop(rect)
        except OverflowError:
            raise MatchResultError(f"Target area({rect}) out of screen{img.size}")
        return target_img

    def _cal_confidence(self, im_source, im_search, rgb):
        """
        Scale the screenshot and the recognition result to the same size, and calculate the reliability

        Args:
             im_source: the image to be matched
             im_search: image template
             rgb: whether to use the rgb channel for verification

        Returns:

        """
        h, w = im_source.size
        im_search = im_search.resize(w, h)
        if rgb:
            confidence = self.template.cal_rgb_confidence(im_source=im_source, im_search=im_search)
        else:
            confidence = self.template.cal_ccoeff_confidence(im_source=im_source, im_search=im_search)

        confidence = (1 + confidence) / 2
        return confidence

    def input_image_check(self, im_source, im_search):
        im_source = self._image_check(im_source)
        im_search = self._image_check(im_search)

        if im_source.place != im_search.place:
            raise InputImageError(
                'image type must be same, source={}, search={}'.format(im_source.place, im_search.place))
        elif im_source.dtype != im_search.dtype:
            raise InputImageError(
                'image data type must be same, source={}, search={}'.format(im_source.dtype, im_search.dtype))
        elif im_source.channels != im_search.channels:
            raise InputImageError(
                'image channel must be same, source={}, search={}'.format(im_source.channels, im_search.channels))

        return im_source, im_search

    def _image_check(self, data):
        if not isinstance(data, Image):
            data = Image(data, dtype=self.Dtype)

        if data.place not in self.Place:
            raise TypeError(f'{self.METHOD_NAME}method,Image type must be(Place.UMat, Place.Ndarray)')
        return data

    @staticmethod
    def _find_homography(sch_pts, src_pts):
        """
        When there are multiple sets of feature point pairs, obtain the unidirectional matrix
        """
        try:
            # M, mask = cv2.findHomography(sch_pts, src_pts, cv2.RANSAC)
            M, mask = cv2.findHomography(sch_pts, src_pts, cv2.RANSAC, 4.0, None, 2000, 0.99)
        except cv2.error:
            import traceback
            traceback.print_exc()
            raise HomographyError("OpenCV error in _find_homography()...")
        else:
            if mask is None:
                raise HomographyError("In _find_homography(), find no mask...")
            else:
                return M, mask

    @staticmethod
    def _perspective_transform(im_source, im_search, src, dst):
        """
        Calculate the perspective transformation according to the
        four pairs of corresponding points, and crop the corresponding picture

        Args:
             im_source: the image to be matched
             im_search: template to be matched
             src: The coordinates of the corresponding quad vertices in the target
                  image (upper left, upper right, lower left, lower right)
             dst: coordinates of the quad vertices in the source image (upper left,
                  upper right, lower left, lower right)

        Returns:

        """
        h, w = im_search.size
        matrix = cv2.getPerspectiveTransform(src=src, dst=dst)
        # warpPerspective https://github.com/opencv/opencv/issues/11784
        output = im_source.warpPerspective(matrix, size=(w, h), flags=cv2.INTER_CUBIC)

        return output

    @staticmethod
    def _get_perspective_area_rect(im_source, src):
        """
        According to the coordinates of the four vertices
        of the rectangle, obtain the largest circumscribed rectangle in the original image

        Args:
            im_source: image to be matched
            src: the coordinates of the corresponding quadrilateral vertices in the target image

        Returns:
            Maximum circumscribed rectangle
        """
        h, w = im_source.size

        x = [int(i[0]) for i in src]
        y = [int(i[1]) for i in src]
        x_min, x_max = min(x), max(x)
        y_min, y_max = min(y), max(y)
        # Selecting the target rectangular area may be out of bounds,
        # and directly set it as the boundary when the boundary is exceeded:
        # If it exceeds the left boundary, it takes 0, if it exceeds the right boundary,
        # it takes w_s-1, if it exceeds the lower boundary, it takes 0, and if it exceeds
        # the upper boundary, it takes h_s-1.
        # When x_min is less than 0, take 0. When x_max is less than 0, take 0.
        x_min, x_max = int(max(x_min, 0)), int(max(x_max, 0))
        # When x_min is greater than w_s, the value is w_s-1. When x_max is greater than w_s-1, take w_s-1。
        x_min, x_max = int(min(x_min, w - 1)), int(min(x_max, w - 1))
        # When y_min is less than 0, take 0. When y_max is less than 0, take 0.
        y_min, y_max = int(max(y_min, 0)), int(max(y_max, 0))
        # When y_min is greater than h_s, the value is h_s-1. When y_max is greater than h_s-1, take h_s-1.
        y_min, y_max = int(min(y_min, h - 1)), int(min(y_max, h - 1))
        rect = Rect(x=x_min, y=y_min, width=(x_max - x_min), height=(y_max - y_min))
        return rect



import base64
import json
import logging
import multiprocessing
from concurrent.futures.thread import ThreadPoolExecutor
from datetime import datetime
from functools import partial
from multiprocessing import Pool
from subprocess import Popen, PIPE
from timeit import default_timer as timer

import flybirds.utils.flybirds_log as log
from flybirds.core.config_manage import WebConfig
from flybirds.utils.dsl_helper import get_use_define_param
from flybirds.utils.uuid_helper import report_name


def create_logger(filename: str):
    """
    create m_logger to dump execution time and duration of features into a
    log file
    :return:
    """
    m_logger = multiprocessing.get_logger()
    m_logger.setLevel(logging.INFO)
    formatter = logging.Formatter(
        '%(asctime)s;%(levelname)s;%(processName)s;%(message)s')
    handler = logging.FileHandler(filename=filename)
    handler.setFormatter(formatter)

    # to avoid duplicated messages in the output
    if not len(m_logger.handlers):
        m_logger.addHandler(handler)
    return m_logger


logger = create_logger('multiprocessing_features.log')


def parallel_run(context):
    """
    Parallel Behave Runner
    """
    behave_cmd = context.get("cmd_str")
    feature_path = context.get("feature_path")

    if behave_cmd is None or feature_path is None:
        raise Exception("[parallel_runner] parse args has error")

    parsed_tags = context.get("parsed_tags")

    if parsed_tags and len(parsed_tags) > 0:
        # -k, --no-skipped
        cmd = f'behave {feature_path} {" ".join(parsed_tags)} -d -k -f json ' \
              f'--no-summary'
    else:
        cmd = f'behave {feature_path} -d -k -f json --no-summary'
    features = get_features_num(cmd)

    log.info('start thread...')
    with ThreadPoolExecutor(max_workers=3) as t_pool:
        browser_types = get_browser_types(context)
        for b_type in browser_types:
            t_pool.submit(multiplication, b_type, context, features)
    log.info('all thread done...')


def multiplication(browser_type, context, features):
    log.info('multiplication start')
    log.info(f'multiplication browser_type:{browser_type}')

    behave_cmd = context.get("cmd_str")
    feature_path = context.get("feature_path")
    processes = context.get("processes")

    cur_browser_type = str(base64.b64encode(browser_type.encode('utf-8')),
                           'utf-8')
    behave_cmd = behave_cmd + f'  -D cur_browser={cur_browser_type}'
    log.info(f'cmd str: {behave_cmd}')

    pool = Pool(processes) if len(features) >= processes else Pool(
        len(features))
    results = pool.map(
        partial(execute_parallel_feature, behave_cmd=behave_cmd,
                feature_path=feature_path, browser_type=browser_type),
        features)
    pool.close()
    pool.join()
    log.info(f'parallel run result: {results}')


def execute_parallel_feature(feature, behave_cmd, feature_path, browser_type):
    """
    Runs features in parallel
    :param feature: feature to run
    :param behave_cmd: behave cmd string
    :param feature_path: feature path
    :param browser_type: browser_type
    """
    feature_start_time = datetime.now()
    start_timer = timer()
    file_name = report_name(feature, browser_type)
    cmd = behave_cmd.replace(feature_path, feature, 1).replace('report.json',
                                                               file_name, 1)
    log.info(f'cmd str: {cmd}')

    p = Popen(cmd, stdout=PIPE, shell=True)
    code = p.wait()
    p.communicate()

    status = 'Passed' if code == 0 else 'Failed'
    logging.info('{0:50}: {1}!!'.format(feature, status))
    feature_end_time = datetime.now()
    end_timer = timer()
    logger.info(f'{feature.split("/")[-1].split(".")[0]};'
                f'{feature_start_time};'
                f'{feature_end_time};'
                f'{end_timer - start_timer};'
                f'{status}')
    return status


def dry_run_parsed_cmd(cmd: str) -> str:
    """
    Execute command and return it's stdout as string
    :param cmd: bash command as str
    :return: stdout from command execution
    """
    p = Popen(cmd, stdout=PIPE, shell=True)
    out, err = p.communicate()

    return json.loads(out.decode('utf-8-sig'))


def get_features_num(cmd: str):
    parsed_output = dry_run_parsed_cmd(cmd)
    if not parsed_output:
        log.warn(
            f'No json output from executed behave dry run command. Nothing to '
            f'be executed.Command: {cmd}\nNothing to execute')
        return
    features = list({feature['location'].split(':')[0]
                     for feature in parsed_output
                     })
    log.info(f'features num need to be executed in parallel: {len(features)}')
    return features


def get_browser_types(context):
    user_data = get_use_define_param(context, 'browserType')
    browser_type = WebConfig(user_data, None).browser_type

    # check browser_type
    browser_types = browser_type
    if isinstance(browser_type, str):
        browser_types = browser_type.split(',')
    browser_types = list(set(browser_types))
    temp = []
    [temp.append(i.strip().lower()) for i in browser_types if
     i.strip().lower() in ['chromium', 'firefox', 'webkit']]
    # add default value
    if len(temp) == 0:
        log.warn(
            'flybirds did not find a browser that would launch. Now chromium '
            'will be launched by default.')
        temp.append('chromium')
    context['browser_types'] = temp
    return temp



# -*- coding: utf-8 -*-
"""
Poco element apis
"""
import time
import os

import flybirds.core.global_resource as gr
import flybirds.core.plugin.plugins.default.ui_driver.poco.findsnap \
    as find_snap
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_manage as pm
import flybirds.utils.flybirds_log as log
from flybirds.core.exceptions import FlybirdEleExistsException
from flybirds.core.exceptions import FlybirdVerifyException
from flybirds.core.global_context import GlobalContext as g_Context
from flybirds.utils import language_helper as lan
from flybirds.core.plugin.plugins.default.step.common import img_verify
from flybirds.core.plugin.plugins.default.step.click import click_image


def wait_exists(poco, selector_str, optional):
    """
    determine whether the element exists within the specified time
    """
    timeout = optional["timeout"]
    context = None
    if 'context' in optional:
        context = optional["context"]
    current_wait_second = 1
    find_success = False
    while timeout > 0:
        create_success = False
        try:
            poco_target = pm.create_poco_object_by_dsl(
                poco, selector_str, optional
            )
            create_success = True
            search_time = current_wait_second
            if search_time > 3:
                search_time = 3
            ele_exists = poco_target.exists()
            log.info(
                "wait_exists: {}, ele_exists: {}, timeout: {}".format(
                    selector_str, ele_exists, timeout
                )
            )

            if ele_exists:
                find_success = True
                break

            poco_target.wait_for_appearance(timeout=search_time)
            find_success = True
            log.info(
                "wait_for_appearance: find_success: {}, timeout: {}".format(
                    find_success, timeout
                )
            )
            break
        except Exception:
            if not create_success:
                time.sleep(1)
        if current_wait_second == 3:
            # modal error detection
            try:
                result = detect_error(context)
                log.info(f"detect_error result:{result}")
                if result is False:
                    break
            except Exception:
                log.info("detect_error exception")
            time.sleep(1)
        if current_wait_second > 4:
            break
        timeout -= current_wait_second
        current_wait_second += 1
    if not find_success:
        message = "during {}s time, not find {} in page".format(
            optional["timeout"], selector_str
        )
        raise FlybirdVerifyException(message)


def not_exist(poco, selector_str, optional):
    """
    determine whether the element does not exist
    """
    ele_exists = False
    try:
        poco_object = pm.create_poco_object_by_dsl(
            poco, selector_str, optional
        )
        ele_exists = poco_object.exists()
    except Exception:
        pass
    if ele_exists:
        message = "{} exists in page".format(selector_str)
        raise FlybirdEleExistsException(message)


def wait_disappear(poco, selector_str, optional):
    """
    determine whether the element disappears within the specified time
    """
    timeout = optional["timeout"]
    current_wait_second = 1
    disappear_success = False
    while timeout > 0:
        create_success = False
        try:
            poco_target = pm.create_poco_object_by_dsl(
                poco, selector_str, optional
            )
            create_success = True
            search_time = current_wait_second
            if search_time > 3:
                search_time = 3
            poco_target.wait_for_disappearance(timeout=search_time)
            disappear_success = True
            break
        except Exception:
            if not create_success:
                time.sleep(1)
        if current_wait_second > 4:
            break
        timeout -= current_wait_second
        current_wait_second += 1
    if not disappear_success:
        message = "during {}s time, {} not disappear in page".format(
            optional["timeout"], selector_str
        )
        raise FlybirdVerifyException(message)


def detect_error(context):
    language = g_Context.get_current_language()
    modal_list = lan.parse_glb_str("modal_list", language)
    break_list = lan.parse_glb_str("break_list", language)
    poco = g_Context.ui_driver_instance

    img_path = "tpl/app"
    if context is not None and os.path.exists(img_path):
        images = sorted([tpl for tpl in os.listdir(img_path) if str(tpl).endswith('png')])
        for img in images:
            result = img_verify(context, os.path.join(img_path, img))
            log.info(f"in detect error method, img detect result is {result}")
            if len(result) > 0:
                click_image(context, img_path)
                log.info("detect_error: x_button_exists: true")
                return True

    for error_str in modal_list:
        log.info(f"in detect error method, error_str detect: {error_str}")
        error_target = pm.create_poco_object_by_dsl(
            poco, error_str, None
        )
        is_existed = error_target.exists()
        if is_existed:
            error_target.click()
            if gr.get_frame_config_value("use_snap", False):
                find_snap.fix_refresh_status(True)
            log.info("detect_error: {}, layer_errors_exists: true"
                     .format(error_str))
            return True

    for break_str in break_list:
        log.info(f"in detect error method, break_str detect: {break_str}")
        break_target = pm.create_poco_object_by_dsl(
            poco, break_str, None
        )
        is_existed = break_target.exists()
        if is_existed:
            break_target.click()
            if gr.get_frame_config_value("use_snap", False):
                find_snap.fix_refresh_status(True)
            log.info("detect_error: {}, layer_errors_exists: true"
                     .format(break_str))
            return False



#! usr/bin/python
# -*- coding:utf-8 -*-
class BaseError(Exception):
    def __init__(self, message="", *args, **kwargs):
        self.message = message

    def __repr__(self):
        return repr(self.message)


class NoEnoughPointsError(BaseError):
    """ detect not enough feature points in input images"""


class HomographyError(BaseError):
    """ An error occurred while findHomography """


class MatchResultError(BaseError):
    """ An error occurred while result out of screen"""


class PerspectiveTransformError(BaseError):
    """ An error occurred while perspectiveTransform """


class InputImageError(BaseError):
    """ An error occurred while input image place/dtype/channels error"""



# -*- coding: utf-8 -*-
"""
when behave start run hook will trigger this
"""

from flybirds.core.global_context import GlobalContext
from flybirds.utils import launch_helper


class OnBefore:  # pylint: disable=too-few-public-methods
    """
    user defined before all hook
    """

    name = "On_Before_User_Hook"
    order = 50

    @staticmethod
    def can(context):
        return True

    @staticmethod
    def run(context):
        """
        user defined before all hook
        """
        # hook extend by tester
        before_all_extend = launch_helper.get_hook_file("before_all_extend")
        if before_all_extend is not None:
            before_all_extend(context)


var = GlobalContext.join("before_run_processor", OnBefore, 1)



# -*- coding: utf-8 -*-
"""
android screen imp
"""
from flybirds.core.plugin.plugins.default.screen import BaseScreen

__open__ = ["Screen"]


class Screen(BaseScreen):
    """
    screen imp
    """
    name = "android_screen"

    # def screen_shot(self, path):
    #     screen.screen_shot(path)
    #
    # def screen_link_to_behave(self, scenario, step_index, tag=None):
    #     screen.screen_link_to_behave(scenario, step_index, tag)



# -*- coding: utf-8 -*-
"""
Element verification
"""
import re
import time

import flybirds.core.global_resource as gr
import flybirds.utils.flybirds_log as log
from flybirds.core.global_context import GlobalContext as g_Context
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_ele \
    as poco_ele
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_findsnap \
    as poco_find_snap
import flybirds.core.plugin.plugins.default.ui_driver.poco.poco_verify \
    as poco_verify
import flybirds.utils.dsl_helper as dsl_helper
import flybirds.utils.verify_helper as verify
from flybirds.core.exceptions import FlybirdVerifyException
from flybirds.core.plugin.plugins.default.step.common import ocr, img_verify


def wait_text_exist(context, param):
    poco_instance = gr.get_value("pocoInstance")

    param_dict = dsl_helper.params_to_dic(param)
    selector_str = param_dict["selector"]
    config = None
    use_snap = gr.get_frame_config_value("use_snap", False)
    if "fuzzyMatch" in param_dict.keys():
        if use_snap:
            config = {"textMatches": selector_str}
        else:
            selector_str = "textMatches=" + selector_str
    else:
        if use_snap:
            config = {"text": selector_str}
        else:
            selector_str = "text=" + selector_str
    optional = {}
    if "timeout" in param_dict.keys():
        optional["timeout"] = float(param_dict["timeout"])
    else:
        optional["timeout"] = gr.get_frame_config_value("wait_ele_timeout", 10)
    if use_snap and config:
        poco_find_snap.find_ele_by_snap(poco_instance, config, optional)
    else:
        poco_ele.wait_exists(poco_instance, selector_str, optional)


def text_not_exist(context, param):
    poco_instance = gr.get_value("pocoInstance")

    param_dict = dsl_helper.params_to_dic(param)
    selector_str = param_dict["selector"]
    if "fuzzyMatch" in param_dict.keys():
        selector_str = "textMatches=" + selector_str
    else:
        selector_str = "text=" + selector_str
    optional = {}

    poco_ele.not_exist(poco_instance, selector_str, optional)


def wait_text_disappear(context, param):
    poco_instance = gr.get_value("pocoInstance")

    param_dict = dsl_helper.params_to_dic(param)
    selector_str = param_dict["selector"]
    if "fuzzyMatch" in param_dict.keys():
        selector_str = "textMatches=" + selector_str
    else:
        selector_str = "text=" + selector_str
    optional = {}
    if "timeout" in param_dict.keys():
        optional["timeout"] = float(param_dict["timeout"])
    else:
        optional["timeout"] = gr.get_frame_config_value(
            "wait_ele_disappear", 10
        )

    poco_ele.wait_disappear(poco_instance, selector_str, optional)


def wait_ele_exit(context, param):
    poco_instance = gr.get_value("pocoInstance")

    param_dict = dsl_helper.params_to_dic(param)
    selector_str = param_dict["selector"]
    optional = {}
    config = None
    use_snap = gr.get_frame_config_value("use_snap", False)
    if "path" in param_dict.keys():
        optional["path"] = param_dict["path"]
    elif "multiSelector" in param_dict.keys():
        optional["multiSelector"] = param_dict["multiSelector"]
    elif use_snap:
        config = {"name": selector_str}
    if "timeout" in param_dict.keys():
        optional["timeout"] = float(param_dict["timeout"])
    else:
        optional["timeout"] = gr.get_frame_config_value("wait_ele_timeout", 10)
    if use_snap and config:
        print("Use snap to determine the existence of elements")
        poco_find_snap.find_ele_by_snap(poco_instance, config, optional)
    else:
        poco_ele.wait_exists(poco_instance, selector_str, optional)


def ele_not_exit(context, param):
    poco_instance = gr.get_value("pocoInstance")
    param_dict = dsl_helper.params_to_dic(param)
    selector_str = param_dict["selector"]
    optional = {}
    if "path" in param_dict.keys():
        optional["path"] = param_dict["path"]
    elif "multiSelector" in param_dict.keys():
        optional["multiSelector"] = param_dict["multiSelector"]

    poco_ele.not_exist(poco_instance, selector_str, optional)


def wait_ele_disappear(context, param):
    poco_instance = gr.get_value("pocoInstance")
    param_dict = dsl_helper.params_to_dic(param)
    selector_str = param_dict["selector"]
    optional = {}
    if "path" in param_dict.keys():
        optional["path"] = param_dict["path"]
    elif "multiSelector" in param_dict.keys():
        optional["multiSelector"] = param_dict["multiSelector"]
    if "timeout" in param_dict.keys():
        optional["timeout"] = float(param_dict["timeout"])
    else:
        optional["timeout"] = gr.get_frame_config_value(
            "wait_ele_disappear", 10
        )

    poco_ele.wait_disappear(poco_instance, selector_str, optional)


def ele_text_equal(context, param1, param2):
    poco_instance = gr.get_value("pocoInstance")
    pattern = re.compile(r"\s+")
    param1_dict = dsl_helper.params_to_dic(param1)
    selector_str = param1_dict["selector"]
    optional = {}
    config = None
    # use snap
    # use_snap = gr.get_frame_config_value("use_snap", False)
    if "path" in param1_dict.keys():
        optional["path"] = param1_dict["path"]
    elif "multiSelector" in param1_dict.keys():
        optional["multiSelector"] = param1_dict["multiSelector"]
    if "timeout" in param1_dict.keys():
        optional["timeout"] = float(param1_dict["timeout"])
    else:
        optional["timeout"] = gr.get_frame_config_value("wait_ele_timeout", 10)

    params_deal_module = None
    deal_method = None
    param2_dict = dsl_helper.params_to_dic(param2, "tText")
    if "dealMethod" in param2_dict.keys():
        deal_method = param2_dict["dealMethod"]
        params_deal_module = gr.get_value("projectScript").params_deal
    target_str = param2_dict["tText"]
    handled_target_str = re.sub(pattern, "", target_str.replace(u"\u200b", ""))
    handled_selector_str = re.sub(
        pattern, "", selector_str.replace(u"\u200b", "")
    )
    if not (deal_method is None):
        deal_method_fun = getattr(params_deal_module, deal_method)
        handled_target_str = deal_method_fun(handled_target_str)
        pattern = re.compile(r"\s+")
        handled_target_str = re.sub(pattern, "", handled_target_str)
    if False:
        config = {"name": selector_str, "expect_text": target_str}
        poco_find_snap.verify_ele_by_snap(poco_instance, config, optional)
    else:
        poco_verify.ele_text_is(
            poco_instance,
            handled_selector_str,
            handled_target_str,
            optional,
            deal_method,
            params_deal_module,
        )


def ele_text_container(context, param1, param2):
    poco_instance = gr.get_value("pocoInstance")

    param1_dict = dsl_helper.params_to_dic(param1)
    selector_str = param1_dict["selector"]
    optional = {}
    if "path" in param1_dict.keys():
        optional["path"] = param1_dict["path"]
    elif "multiSelector" in param1_dict.keys():
        optional["multiSelector"] = param1_dict["multiSelector"]
    if "timeout" in param1_dict.keys():
        optional["timeout"] = float(param1_dict["timeout"])
    else:
        optional["timeout"] = gr.get_frame_config_value("wait_ele_timeout", 10)

    deal_method = None
    params_deal_module = None
    param2_dict = dsl_helper.params_to_dic(param2, "tText")
    if "dealMethod" in param2_dict.keys():
        deal_method = param2_dict["dealMethod"]
        params_deal_module = gr.get_value("projectScript").params_deal
    target_str = param2_dict["tText"]
    pattern = re.compile(r"\s+")
    handled_target_str = re.sub(pattern, "", target_str.replace(u"\u200b", ""))
    if not (deal_method is None):
        deal_method_fun = getattr(params_deal_module, deal_method)
        handled_target_str = deal_method_fun(handled_target_str)
        pattern = re.compile(r"\s+")
        handled_target_str = re.sub(pattern, "", handled_target_str)

    poco_verify.ele_text_contains(
        poco_instance,
        selector_str,
        handled_target_str,
        optional,
        deal_method,
        params_deal_module,
    )


def wait_ele_appear(context, param):
    poco_instance = gr.get_value("pocoInstance")

    param_dict = dsl_helper.params_to_dic(param)
    selector_str = param_dict["selector"]
    optional = {}
    if "path" in param_dict.keys():
        optional["path"] = param_dict["path"]
    elif "multiSelector" in param_dict.keys():
        optional["multiSelector"] = param_dict["multiSelector"]
    if "timeout" in param_dict.keys():
        optional["timeout"] = float(param_dict["timeout"])
    else:
        optional["timeout"] = gr.get_frame_config_value(
            "page_render_timeout", 30
        )
    optional["context"] = context
    poco_ele.wait_exists(poco_instance, selector_str, optional)


def exist_ele(context, param):
    """
    Compatible with the old version of the statement, it will be discarded in
    the future.
    """
    poco_instance = gr.get_value("pocoInstance")

    params_array = param.split(",")
    selector_str = params_array[0]
    optional = {}
    if len(params_array) >= 2:
        optional["timeout"] = float(params_array[1])
    else:
        optional["timeout"] = gr.get_frame_config_value(
            "wait_ele_timeout", 10
        )
    optional["context"] = context
    poco_ele.wait_exists(poco_instance, selector_str, optional)


def wait_ocr_text_appear(context, param):
    timeout = gr.get_frame_config_value("page_render_timeout", 30)
    start = time.time()
    while True:
        ocr(context)
        txts = [line[1][0] for line in g_Context.ocr_result]
        fixed_txt = paddle_fix_txt(txts, True)
        trim_param = param.replace(" ", "")
        for txt in fixed_txt:
            if trim_param in txt:
                log.info(f"[ocr txt contain] param: {param} found in txt: {txt}")
                return True
            try:
                if re.search(param, txt, flags=0) is not None:
                    log.info(f"[ocr txt contain re] param: {param} found in txt: {txt}")
                    return True
            except:
                pass
            line_param = trim_param.replace("-", "")
            line_txt = txt.replace("-", "")
            if line_param in line_txt:
                log.warn(f"[ocr txt contain line replace] param: {param} found in txt: {txt}")
                return True
        if time.time() - start > timeout:
            for line in g_Context.ocr_result:
                log.info(f"[image ocr result] scan line info is:{line}")
            message = "text not found in {} seconds, expect text:{}" \
                .format(timeout, param)
            raise FlybirdVerifyException(message)


def ocr_txt_exist(context, param):
    if len(g_Context.ocr_result) >= 1:
        txts = [line[1][0] for line in g_Context.ocr_result]
        fixed_txt = paddle_fix_txt(txts, True)
        trim_param = param.replace(" ", "")
        verify.text_container(trim_param, fixed_txt)
    else:
        for line in g_Context.ocr_result:
            log.info(f"[image ocr result] scan line info is:{line}")
        message = "ocr result is null"
        raise FlybirdVerifyException(message)


def ocr_txt_contain(context, param, islog=True):
    if len(g_Context.ocr_result) >= 1:
        txts = [line[1][0] for line in g_Context.ocr_result]
        fixed_txt = paddle_fix_txt(txts, True)
        trim_param = param.replace(" ", "")
        for txt in fixed_txt:
            if trim_param in txt:
                log.info(f"[ocr txt contain] param: {param} found in txt: {txt}")
                return True
            try:
                if re.search(param, txt, flags=0) is not None:
                    log.info(f"[ocr txt contain re] param: {param} found in txt: {txt}")
                    return True
            except:
                pass
            line_param = trim_param.replace("-", "")
            line_txt = txt.replace("-", "")
            if line_param in line_txt:
                log.warn(f"[ocr txt contain line replace] param: {param} found in txt: {txt}")
                return True
        if islog is True:
            for line in g_Context.ocr_result:
                log.info(f"[image ocr result] scan line info is:{line}")
        message = "ocr result not contain {}".format(param)
        raise FlybirdVerifyException(message)
    else:
        message = "[ocr txt contain] ocr result is null"
        raise FlybirdVerifyException(message)


def ocr_txt_not_exist(context, param):
    if len(g_Context.ocr_result) >= 1:
        txts = [line[1][0] for line in g_Context.ocr_result]
        fixed_txt = paddle_fix_txt(txts, True)
        trim_param = param.replace(" ", "")
        verify.text_not_container(trim_param, fixed_txt)
    else:
        for line in g_Context.ocr_result:
            log.info(f"[image ocr result] scan line info is:{line}")
        message = "ocr result is null"
        raise FlybirdVerifyException(message)


def paddle_fix_txt(txt, trim=False):
    # replace value in paddle_fix config
    paddle_fix = gr.get_paddle_fix_value()
    if paddle_fix is not None:
        for i in range(len(txt)):
            for key in paddle_fix:
                origin_txt = txt[i]
                if key in origin_txt:
                    txt[i] = origin_txt.replace(key, paddle_fix[key])
    # replace  blank value
    if trim is True:
        for i in range(len(txt)):
            origin_txt = txt[i]
            txt[i] = origin_txt.replace(" ", "")
    return txt


def img_exist(context, param, islog=True):
    start = time.time()
    step_index = context.cur_step_index - 1
    result = img_verify(context, param)
    if len(result) == 0:
        if islog is True:
            src_path = "../../../{}".format(param)
            data = (
                'embeddingsTags, stepIndex={}, <image class ="screenshot"'
                ' width="375" src="{}" />'.format(step_index, src_path)
            )
            context.scenario.description.append(data)
        raise Exception("[image exist verify] image not found !")
    else:
        log.info(f"[image exist verify] cost time:{time.time() - start}")
        log.info(f"[image exist verify] result:{result}")
        return True


def img_not_exist(context, param):
    start = time.time()
    step_index = context.cur_step_index - 1
    result = img_verify(context, param)
    if len(result) == 0:
        log.info(f"[image not exist verify] cost time:{time.time() - start}")
        log.info(f"[image not exist verify] result:{result}")
    else:
        src_path = "../../../{}".format(param)
        data = (
            'embeddingsTags, stepIndex={}, <image class ="screenshot"'
            ' width="375" src="{}" />'.format(step_index, src_path)
        )
        context.scenario.description.append(data)
        # context.cur_step_index += 1
        raise Exception("[image not exist verify] image found !")



# -*- coding: utf-8 -*-
"""
android Element core api implement
"""
from flybirds.core.plugin.plugins.default.base_element import BaseElement

__open__ = ["Element"]


class Element(BaseElement):
    """Android Element Class"""

    name = "android_element"

    def ui_driver_init(self):
        """
        Initialize the poco object
         :return:
        """
        from poco.drivers.android.uiautomation import AndroidUiautomationPoco

        poco = AndroidUiautomationPoco(
            use_airtest_input=True, screenshot_each_action=False
        )
        return poco



# -*- coding: utf-8 -*-
"""
Convert selector
"""
from flybirds.core.global_context import GlobalContext as g_Context
from flybirds.utils import language_helper as lan


def create_multi_selector(multi_str, default="name"):
    """
    description of the poco multi-condition selection element
    is converted into a dictionary for use by poco
    :param multi_str:
    :param default:
    :return:
    """
    # get current language
    language = g_Context.get_current_language()
    split_str = lan.parse_glb_str("and", language)

    selector_array = multi_str.split(split_str)
    result = {}
    for item in selector_array:
        if "=" in item:
            item_split = item.split("=", 1)
            key = item_split[0].strip()
            result[key] = item_split[1].strip()
        else:
            result[default] = item.strip()
    return result


def create_single_selector(single_str, default="name"):
    """
    description of the poco single condition selection element
    is converted into a dictionary for use by poco
    """
    result = {}
    if "=" in single_str:
        single_split = single_str.split("=", 1)
        key = single_split[0].strip()
        result[key] = single_split[1].strip()
    else:
        result[default] = single_str.strip()
    return result



# -*- coding: utf-8 -*-
"""
This module defines the steps related to the page.
"""

from behave import step

from flybirds.core.global_context import GlobalContext as g_Context
from flybirds.utils.dsl_helper import ele_wrap


@step("go to url[{param}]")
@ele_wrap
def jump_to_page(context, param=None):
    g_Context.step.jump_to_page(context, param)


@step("return to previous page")
def return_pre_page(context):
    g_Context.step.return_pre_page(context)


@step("go to home page")
def to_app_home(context):
    g_Context.step.to_app_home(context)


@step("logon account[{selector1}]password[{selector2}]")
@ele_wrap
def app_login(context, selector1=None, selector2=None):
    g_Context.step.app_login(context, selector1, selector2)


@step("logout")
def app_logout(context):
    g_Context.step.app_logout(context)


@step("unblock the current page")
def unblock_page(context):
    g_Context.step.unblock_page(context)


@step("current page is [{param}]")
@ele_wrap
def cur_page_is(context, param=None):
    g_Context.step.cur_page_is(context, param)


@step("current page is not last page")
def has_page_changed(context):
    g_Context.step.has_page_changed(context)



# -*- coding: utf-8 -*-
"""
load event
"""
import base64
import os

from flybirds.utils.pkg_helper import load_pkg_by_ns

import flybirds.core.global_resource as gr
import flybirds.utils.flybirds_log as log
from flybirds.utils.flybirds_log import logger
from flybirds.core.config_manage import DeviceConfig
from flybirds.core.config_manage import PluginConfig
from flybirds.core.global_context import GlobalContext
from flybirds.core.plugin.plugin_manager import DirectoryPluginManager
from flybirds.utils.dsl_helper import str2bool


class PluginManager:  # pylint: disable=too-few-public-methods
    """
    plugin load management
    """

    name = "PluginManager"
    order = 1

    @staticmethod
    def run(context):
        """
        load plugin from pkg
        """
        # initialize the global object module
        gr.init_glb()
        # get user-specified parameter value or custom parameter
        user_data = {}
        if context is not None and context.config is not None:
            user_data = (
                context.config.userdata
                if isinstance(context.config.userdata, dict)
                else {}
            )
            for key, value in user_data.items():
                user_data[key] = str(base64.b64decode(value), "utf-8")

        # check some specific values of the user_data
        if user_data.get('platform'):
            platform = user_data.get('platform').strip().lower()
            if platform not in ['ios', 'android', 'web']:
                log.warn(f'flybirds is not supports to run on {platform} '
                         f'platform. It will now run on Android by default.')
                platform = "android"
            user_data['platform'] = platform

        if user_data.get('headless'):
            headless = str2bool(user_data.get('headless').strip())
            user_data['headless'] = headless

        if user_data.get('browserType'):
            browser_type = user_data.get('browserType').strip().lower()
            browser_types = browser_type.split(',')
            browser_types = list(set(browser_types))
            temp = []
            [temp.append(i.strip().lower()) for i in browser_types if
             i.strip().lower() in ['chromium', 'firefox', 'webkit']]
            if len(temp) == 0:
                temp.append('chromium')
            user_data['browserType'] = ",".join(temp)

        log.info(f'[loader] user_data: {user_data}')
        gr.set_value("userData", user_data)

        cur_browser = "chromium"
        if user_data.get('cur_browser'):
            cur_browser = user_data.get('cur_browser')
        gr.set_value("cur_browser", cur_browser)

        p_info = PluginConfig(user_data).plugin_info
        if p_info is not None and p_info.__contains__("active"):
            GlobalContext.active_plugin = p_info.get("active")

        GlobalContext.plugin_info = p_info.get(GlobalContext.active_plugin)

        if GlobalContext.plugin_info is None:
            raise Exception(
                f"not exist this plugin {GlobalContext.active_plugin}"
            )
        GlobalContext.platform = DeviceConfig(user_data, None).platform
        log.info(
            f"[loader] run platform: {GlobalContext.platform}")
        plugin_manager = DirectoryPluginManager()
        plugin_manager.load_plugins()

        return plugin_manager


class ExtendPluginLoader:
    """
    extend plugin event load
    """
    name = "ExtendPluginLoader"
    order = 20

    @staticmethod
    def run(context):
        """
        load extend hook
        """
        try:
            if os.environ.get("extend_pkg_list") is not None:
                extend_pkg = os.environ.get("extend_pkg_list")
                extend_pkg_list = extend_pkg.split(",")
                if len(extend_pkg_list) > 0:
                    for pkg in extend_pkg_list:
                        if pkg is not None and pkg != "":
                            logger.info(f"load extend hook package:{pkg}")
                            load_pkg_by_ns(f"{pkg}.hook")
        except Exception as load_ex:
            logger.debug(f"no extend hook to be loaded--{load_ex}")


# add plugin load event to global processor
GlobalContext.join("plugin_processor", PluginManager, 1)

GlobalContext.join("plugin_processor", ExtendPluginLoader, 1)



# -*- coding: utf-8 -*-
"""
app file prepare
"""
import os
import flybirds.core.global_resource as gr
import flybirds.utils.flybirds_log as log
from flybirds.core.global_context import GlobalContext
from flybirds.utils import file_helper
from flybirds.utils import download_helper


class OnPrepare:
    """
    prepare
    """
    name = "OnFilePrepare"
    order = 20
    _app_name = "installer.apk"

    @staticmethod
    def check_config():
        """
        only local and has package path
        """
        run_info = gr.get_run_info()
        has_local = False
        has_pkg_path = False
        is_rerun = False
        if run_info is not None and run_info.run_at is not None \
                and run_info.run_at.lower() == "local":
            has_local = True

        if run_info is not None and run_info.is_rerun is not None:
            is_rerun = run_info.is_rerun

        package_path = gr.get_app_config_value("package_path")
        if package_path is not None and package_path.strip() != "":
            has_pkg_path = True
        else:
            log.info("not have package path")
            has_pkg_path = False
        if has_local is True and has_pkg_path is True and is_rerun is False:
            return True
        else:
            return False

    @staticmethod
    def is_web_url(package_path):
        return package_path.startswith(("http:", "https:"))

    @staticmethod
    def check_apk_file(context):
        package_path = gr.get_app_config_value("package_path")
        if OnPrepare.is_web_url(package_path):
            d_apk_path = os.path.join(os.getcwd(), "download",
                                      OnPrepare._app_name)
            if os.path.exists(d_apk_path):
                context.installer_path = d_apk_path
                return False
            else:
                return True
        else:
            if os.path.isfile(package_path):
                if os.path.exists(package_path):
                    context.installer_path = package_path
                    return False
                else:
                    raise Exception(
                        f"cannot find your package in path:{package_path}")
            else:
                new_package_path = os.path.join(package_path,
                                                OnPrepare._app_name)
                if os.path.exists(new_package_path):
                    context.installer_path = new_package_path
                    return False
                else:
                    raise Exception(
                        f"cannot find your package in path:{new_package_path}")

    @staticmethod
    def can(context):
        """
        only native android can be run
        """
        if gr.get_platform() is not None and gr.get_platform().lower() \
                == "android":
            # check config
            config_check = OnPrepare.check_config()
            if config_check is True:
                return OnPrepare.check_apk_file(context)
            else:
                return False
        else:
            return False

    @staticmethod
    def run(context):
        """
        install app
        """
        # download app
        log.info("install file prepare")

        package_path = gr.get_app_config_value("package_path")
        log.info(f"package path:{package_path}")
        installer = None
        if OnPrepare.is_web_url(package_path):
            base_path = os.path.join(os.getcwd(), "download")
            file_helper.create_dirs(base_path)
            d_apk_path = os.path.join(os.getcwd(), "download",
                                      OnPrepare._app_name)
            log.info("download start")
            download_helper.downlaod(package_path, d_apk_path)
            log.info("download end")
            installer = d_apk_path
        else:
            if os.path.isfile(package_path):
                installer = package_path
            else:
                installer = os.path.join(package_path,
                                         OnPrepare._app_name)
        if os.path.exists(installer) is not True:
            raise Exception(
                f"not exist path :{installer}")
        else:
            context.installer_path = installer


# add android prepare into global context
var = GlobalContext.join("before_run_processor", OnPrepare, 1)



# -*- coding: utf-8 -*-
"""
android screen record
"""
from flybirds.core.plugin.plugins.default.screen_record import ScreenRecord

__open__ = ["ScreenRecordInfo"]


class ScreenRecordInfo(ScreenRecord):
    name = "android_screen_record"
    instantiation_timing = "plugin"

    def __init__(self):
        ScreenRecord.__init__(self)



# -*- coding: utf-8 -*-
"""
android step implements class
"""

from flybirds.core.plugin.plugins.default.app_base_step import AppBaseStep
from flybirds.core.plugin.plugins.default.step.app \
    import (install_app, uninstall_app, return_pre_page)
from flybirds.core.plugin.plugins.default.step.schema import jump_to_page

__open__ = ["Step"]


class Step(AppBaseStep):
    """Android Step Class"""

    name = "android_step"

    def install_app(self, context, param):
        install_app(context, param)

    def uninstall_app(self, context, param):
        uninstall_app(context, param)

    def return_pre_page(self, context):
        return_pre_page(context)

    def jump_to_page(self, context, param):
        jump_to_page(context, param)
